from typing import Dict, List
import numbers
from etk.knowledge_graph_schema import KGSchema
from etk.field_types import FieldType
from etk.etk_exceptions import KgValueError, UndefinedFieldError
from etk.knowledge_graph_provenance_record import KnowledgeGraphProvenanceRecord
from etk.extraction import Extraction
from etk.segment import Segment
from etk.ontology_api import Ontology
import json


class KnowledgeGraph(object):
    """
    This class is a knowledge graph object, provides API for user to construct their kg.
    Add field and value to the kg object, analysis on provenance
    """

    def __init__(self, schema: KGSchema, ontology: Ontology, doc) -> None:
        self._kg = {}
        self.origin_doc = doc
        self.schema = schema
        self.ontology = ontology
        if self.origin_doc.etk.generate_json_ld:
            if "doc_id" in doc.cdr_document:
                self.add_value("@id", self.origin_doc.cdr_document["doc_id"])

    def validate_field(self, field_name: str) -> bool:
        """
        Complain if a field in not in the schema
        Args:
            field_name:

        Returns: True if the field is present.

        """
        if field_name in {"@id", "@type"}:
            return True
        result = self.schema.has_field(field_name)
        if not result:
            # todo: how to comply with our error handling policies?
            raise UndefinedFieldError("'{}' should be present in the knowledge graph schema.".format(field_name))
        return result

    def _add_single_value(self, field_name: str, value, provenance_path=None,
                          reference_type="location", keep_empty: bool = False) -> bool:

        def need_to_parse(v):
            if v is not None:  # can not be none
                if isinstance(v, str):  # type is string, only keep it if it's not "" or keep_empty is True
                    if v.strip() != "" or keep_empty:
                        return True
                    else:
                        return False
                else:  # type is not string, keep it
                    return True
            return False

        if field_name == "@id":
            self._kg["@id"] = value
            return True
        if field_name == "@type" and self.origin_doc.etk.generate_json_ld:
            self._kg["@type"] = self._kg.get("@type", list())
            if value not in self._kg["@type"]:
                self._kg["@type"].append(value)
            return True

        if not need_to_parse(value):
            # here, if the value is "empty", then it doesn't need to do further parsing anymore
            # but it's still valid, then True can be returned
            return True

        (valid, this_value) = self.schema.is_valid(field_name, value)
        if self.ontology and self.origin_doc.etk.generate_json_ld:
            # Check and generate the valid value again from 'value'. this_value contains serialized results
            # and since we need toc heck the datatype again, we shouldnt use that.
            valid_value = valid and self.ontology.is_valid(field_name, value, self._kg)
            this_value = valid_value
        else:
            valid_value = valid and {'value': this_value, 'key': self.create_key_from_value(this_value, field_name)}
        if valid_value:
            if valid_value not in self._kg[field_name]:
                self._kg[field_name].append(valid_value)
            self.create_kg_provenance(reference_type, str(this_value), provenance_path) \
                if provenance_path else self.create_kg_provenance(reference_type, str(this_value))
            return True
        else:
            return False

    def _add_value(self, field_name: str, value, provenance_path=None) -> bool:
        """
        Helper function to add values to a knowledge graph
        Args:
            field_name: a field in the knowledge graph, assumed correct
            value: any Python type

        Returns: True if the value is compliant with the field schema, False otherwise

        """
        if not isinstance(value, list):
            value = [value]

        all_valid = True
        for x in value:
            valid = self._add_single_value(field_name, x, provenance_path=provenance_path)
            all_valid = all_valid and valid
        return all_valid

    def _add_doc_value(self, field_name: str, jsonpath: str) -> None:
        """
        Add a value to knowledge graph by giving a jsonpath

        Args:
            field_name: str
            jsonpath: str

        Returns:
        """
        path = self.origin_doc.etk.parse_json_path(jsonpath)
        matches = path.find(self.origin_doc.value)
        all_valid = True
        invalid = []
        for a_match in matches:
            # If the value is the empty string, we treat is a None.
            if a_match.value:
                valid = self._add_value(field_name, a_match.value, provenance_path=str(a_match.full_path))
                if not valid:
                    invalid.append(field_name + ":" + str(a_match.value))
                all_valid = all_valid and valid

        if not all_valid:
            raise KgValueError("Some kg value type invalid according to schema: " + json.dumps(invalid))

    def add_value(self, field_name: str, value: object = None, json_path: str = None,
                  json_path_extraction: str = None, keep_empty: bool = False) -> None:
        """
        Add a value to knowledge graph.
        Input can either be a value or a json_path. If the input is json_path, the helper function _add_doc_value is
        called.
        If the input is a value, then it is handled

        Args:
            field_name: str, the field name in the knowledge graph
            value: the value to be added to the knowledge graph
            json_path: str, if json_path is provided, then get the value at this path in the doc
            json_path_extraction: str,
            discard_empty: bool,
        Returns:
        """

        def validate(v):
            if v is not None:
                if isinstance(v, str):
                    if v.strip() != "" or keep_empty:
                        return True
                    else:
                        return False
                else:
                    return True
            return False

        self.validate_field(field_name)
        if field_name not in self._kg:
            self._kg[field_name] = []

        if json_path:
            self._add_doc_value(field_name, json_path)

        if validate(value):
            if not isinstance(value, list):
                value = [value]

            all_valid = True
            invalid = []
            for a_value in value:
                if isinstance(a_value, Extraction):
                    valid = self._add_single_value(field_name, a_value.value, provenance_path=str(json_path_extraction),
                                                   keep_empty=keep_empty)
                elif isinstance(a_value, Segment):
                    valid = self._add_single_value(field_name, a_value.value, provenance_path=a_value.json_path,
                                                   keep_empty=keep_empty)
                else:
                    valid = self._add_single_value(field_name, a_value, provenance_path=json_path_extraction,
                                                   reference_type="constant", keep_empty=keep_empty)

                all_valid = all_valid and valid
                if not valid:
                    invalid.append(field_name + ":" + str(a_value))
            if not all_valid:
                print("Some kg value type invalid according to schema:" + json.dumps(invalid))
                # raise KgValueError("Some kg value type invalid according to schema")
        # IF we did not add any value, remove the empty field we just added to kg
        if len(self._kg[field_name]) == 0:
            self._kg.pop(field_name)

    @property
    def value(self) -> Dict:
        """
        Get knowledge graph object

        Args:

        Returns: Dict
        """
        return self._kg

    def get_values(self, field_name: str) -> List[object]:
        """
        Get a list of all the values of a field.

        Args:
            field_name:

        Returns: the list of values (not the keys)

        """
        result = list()
        if self.validate_field(field_name):
            for value_key in self._kg.get(field_name):
                result.append(value_key["value"])
        return result

    def create_key_from_value(self, value, field_name: str):
        """

        Args:
            value:
            field_name: str
        Returns: key

        """
        key = value
        if self.schema.field_type(field_name) == FieldType.KG_ID:
            pass
        elif (isinstance(key, str) or isinstance(key, numbers.Number)) and self.schema.field_type(
                field_name) != FieldType.DATE:
            # try except block because unicode characters will not be lowered
            try:
                key = str(key).strip().lower()
            except:
                pass

        return key

    def create_kg_provenance(self, reference_type, value, json_path: str = None) -> None:
        new_id = self.origin_doc.provenance_id_index
        kg_provenance_record: KnowledgeGraphProvenanceRecord = KnowledgeGraphProvenanceRecord(new_id,
                                                                                              "kg_provenance_record",
                                                                                              reference_type, value,
                                                                                              json_path,
                                                                                              self.origin_doc)
        self.origin_doc.provenance_id_index_incrementer()
        if value in self.origin_doc.kg_provenances:
            self.origin_doc.kg_provenances[value].append(new_id)
        else:
            self.origin_doc.kg_provenances[value] = [new_id]
        self.origin_doc.provenances[new_id] = kg_provenance_record
        if "provenances" not in self.origin_doc.cdr_document:
            self.origin_doc.cdr_document["provenances"] = []
        _dict = self.get_dict_kg_provenance(kg_provenance_record)
        self.origin_doc.cdr_document["provenances"].append(_dict)

    @staticmethod
    def get_dict_kg_provenance(kg_provenance_record: KnowledgeGraphProvenanceRecord):
        _dict = dict()
        _dict["@id"] = kg_provenance_record.id
        _dict["@type"] = kg_provenance_record._type
        _dict["reference_type"] = kg_provenance_record.reference_type
        _dict["value"] = kg_provenance_record._value
        if kg_provenance_record.json_path is not None:
            _dict["json_path"] = kg_provenance_record.json_path
        return _dict

    def context_resolve(self, field_uri: str) -> str:
        """
        According to field_uri to add corresponding context and return a resolvable field_name

        :param field_uri:
        :return: a field_name that can be resolved with kg's @context
        """
        from rdflib.namespace import split_uri
        context = self._kg["@context"] = self._kg.get("@context", dict())
        nm = self.ontology.g.namespace_manager
        space, name = split_uri(field_uri)
        if "@vocab" not in context and None in nm.namespaces():
            context["@vocab"] = nm.store.prefix(space)
        if "@vocab" in context and space == context["@vocab"]:
            # case #1, can directly use name
            return name
        if self.schema.has_field(name):
            if name not in context:
                prefix = [x for x in list(self.ontology.g.namespace_manager.namespaces())]
                for x, y in prefix:
                    if space[:-1] == x:
                        context[name] = str(y) + name
                        return name
                context[name] = field_uri
            return name
        prefix = nm.store.prefix(space)
        if prefix:
            context[prefix] = space
            return nm.qname(field_uri)
        return field_uri



from etk.origin_record import OriginRecord
from functools import reduce
import itertools


class ProvenanceAPI:

    def __init__(self, doc):
        self.origin_doc = doc

    def get_origins(self, field: str = None, index: int = None, value: str = None):
        origins = []
        if value == None:
            if index == None:
                items = self.origin_doc.kg.value[field]
                for item in items:
                    search_string = item["value"]
                    kg_provenance_ids = self.origin_doc.kg_provenances[search_string]  # .get_origins(search_string)

                    for kg_provenance_id in kg_provenance_ids:
                        kg_provenance = self.origin_doc.provenances[kg_provenance_id]
                        origins.extend(kg_provenance.get_origins(search_string))
            else:
                search_string = self.origin_doc.kg.value[field][index]["value"]
                kg_provenance_ids = self.origin_doc.kg_provenances[search_string]  # .get_origins(search_string)
                for kg_provenance_id in kg_provenance_ids:
                    kg_provenance = self.origin_doc.provenances[kg_provenance_id]
                    origins.extend(kg_provenance.get_origins(search_string))

        else:
            kg_provenance_ids = self.origin_doc.kg_provenances[value]  # .get_origins(search_string)
            for kg_provenance_id in kg_provenance_ids:
                kg_provenance = self.origin_doc.provenances[kg_provenance_id]
                origins.extend(kg_provenance.get_origins(value))
        return origins

# TODO: implement the method below
    # def get_methods(self, field: str = None, index: int = None, value: str = None):
    #     search_string = self.origin_doc.kg.value[field][index]["value"]
    #     methods = kg_provenance.get_methods(search_string)
    #     return methods



import unittest
import json
from etk.extractors.html_metadata_extractor import HTMLMetadataExtractor


class TestMetadataExtractor(unittest.TestCase):
    # value of title extraction is a string
    def test_title_extractor(self) -> None:
        hme = HTMLMetadataExtractor()
        with open('etk/unit_tests/ground_truth/sample_html.jl', 'r') as f:
            sample_html = json.load(f)
        test_html = sample_html["raw_content"]

        title_text = hme.extract(test_html, extract_title = True)[0].value
        expected_title = "323-452-2013 ESCORT ALERT! - Luna The Hot Playmate (323) 452-2013 - 23"
        self.assertEqual(title_text, expected_title)

    # value of mate extraction is a dict
    def test_meta_tag_extractor(self) -> None:
        hme = HTMLMetadataExtractor()
        with open('etk/unit_tests/ground_truth/sample_html.jl', 'r') as f:
            sample_html = json.load(f)
        test_html = sample_html["raw_content"]

        meta_tag_dic = hme.extract(test_html, extract_meta = True)[0].value
        expected_dic = {
                'description': "323-452-2013 Escort listings. View all of her listings at once. On this page you will see a history for 323-452-2013 and all of the escort ads placed with this phone number. Specifically in losangeles: 323-452-2013 ESCORT ALERT! - Luna The Hot Playmate (323) 452-2013 - 23",
                'netinsert': "0.0.1.13.13.1",
                'keywords': "323-452-2013, escort listing, escort ad, escorts in Los angeles, 323-452-2013 in Los angeles, escorts in losangeles, call 323-452-2013, phone number 323-452-2013, craigslist escorts, backpage escorts, listings"
            }

        self.assertDictEqual(meta_tag_dic, expected_dic)

    # value of microdata extraction is a list
    def test_microdata_extractor(self) -> None:
        hme = HTMLMetadataExtractor()
        test_html = """<!DOCTYPE HTML>
                    ... <html>
                    ...  <head>
                    ...   <title>Photo gallery</title>
                    ...  </head>
                    ...  <body>
                    ...   <h1>My photos</h1>
                    ...   <figure itemscope itemtype="http://n.whatwg.org/work" itemref="licenses">
                    ...    <img itemprop="work" src="images/house.jpeg" alt="A white house, boarded up, sits in a forest.">
                    ...    <figcaption itemprop="title">The house I found.</figcaption>
                    ...   </figure>
                    ...   <figure itemscope itemtype="http://n.whatwg.org/work" itemref="licenses">
                    ...    <img itemprop="work" src="images/mailbox.jpeg" alt="Outside the house is a mailbox. It has a leaflet inside.">
                    ...    <figcaption itemprop="title">The mailbox.</figcaption>
                    ...   </figure>
                    ...   <footer>
                    ...    <p id="licenses">All images licensed under the <a itemprop="license"
                    ...    href="http://www.opensource.org/licenses/mit-license.php">MIT
                    ...    license</a>.</p>
                    ...   </footer>
                    ...  </body>
                    ... </html>"""

        microdata = hme.extract(test_html, extract_microdata = True)[0].value
        expected_microdata = [{'properties': {'license': 'http://www.opensource.org/licenses/mit-license.php',
                                             'title': 'The house I found.',
                                             'work': 'http://www.example.com/images/house.jpeg'},
                                            'type': 'http://n.whatwg.org/work'},
                                            {'properties': {'license': 'http://www.opensource.org/licenses/mit-license.php',
                                             'title': 'The mailbox.',
                                             'work': 'http://www.example.com/images/mailbox.jpeg'},
                                            'type': 'http://n.whatwg.org/work'}]
        self.assertEqual(microdata, expected_microdata)

    # value of microdata extraction is list
    def test_json_ld_extractor(self) -> None:
        hme = HTMLMetadataExtractor()
        with open('etk/unit_tests/ground_truth/sample_html_for_json_ld.html', 'r') as f:
          test_html = f.read()
        
        json_ld_data = hme.extract(test_html, extract_json_ld = True)[0].value

        expected_ld_data = [{'@context': 'http://schema.org',
                          '@type': 'MusicEvent',
                          'location': {'@type': 'Place',
                                       'address': {'@type': 'PostalAddress',
                                                   'addressCountry': 'US',
                                                   'addressLocality': 'Brooklyn',
                                                   'addressRegion': 'NY',
                                                   'postalCode': '11225',
                                                   'streetAddress': '497 Rogers Ave'},
                                       'geo': {'@type': 'GeoCoordinates',
                                               'latitude': 40.660109,
                                               'longitude': -73.953193},
                                       'name': 'The Owl Music Parlor',
                                       'sameAs': 'http://www.theowl.nyc'},
                          'name': 'Elysian Fields',
                          'performer': [{'@type': 'MusicGroup',
                                         'name': 'Elysian Fields',
                                         'sameAs': 'https://www.songkick.com/artists/236156-elysian-fields?utm_medium=organic&utm_source=microformat'}],
                          'startDate': '2017-06-10T19:30:00-0400',
                          'url': 'https://www.songkick.com/concerts/30173984-elysian-fields-at-owl-music-parlor?utm_medium=organic&utm_source=microformat'},
                         {'@context': 'http://schema.org',
                          '@type': 'MusicEvent',
                          'location': {'@type': 'Place',
                                       'address': {'@type': 'PostalAddress',
                                                   'addressCountry': 'US',
                                                   'addressLocality': 'San Francisco',
                                                   'addressRegion': 'CA',
                                                   'postalCode': '94107',
                                                   'streetAddress': '500 Fourth Street'},
                                       'geo': {'@type': 'GeoCoordinates',
                                               'latitude': 37.7795638,
                                               'longitude': -122.398023},
                                       'name': 'Hotel Utah Saloon',
                                       'sameAs': 'http://www.hotelutah.com/'},
                          'name': 'Elysian Fields',
                          'performer': [{'@type': 'MusicGroup',
                                         'name': 'Elysian Fields',
                                         'sameAs': 'https://www.songkick.com/artists/236156-elysian-fields?utm_medium=organic&utm_source=microformat'},
                                        {'@type': 'MusicGroup',
                                         'name': 'Chocolate Genius Inc.',
                                         'sameAs': 'https://www.songkick.com/artists/1009602-chocolate-genius-inc?utm_medium=organic&utm_source=microformat'}],
                          'startDate': '2017-04-26T20:00:00-0700',
                          'url': 'https://www.songkick.com/concerts/29673614-elysian-fields-at-hotel-utah-saloon?utm_medium=organic&utm_source=microformat'},
                         {'@context': 'http://schema.org',
                          '@type': 'MusicEvent',
                          'location': {'@type': 'Place',
                                       'address': {'@type': 'PostalAddress',
                                                   'addressCountry': 'France',
                                                   'addressLocality': 'Saint-Nazaire',
                                                   'postalCode': '44600',
                                                   'streetAddress': 'Alvéole 14 de la base '
                                                                    'sous-Marine Bd de la '
                                                                    'Légion d’Honneur'},
                                       'geo': {'@type': 'GeoCoordinates',
                                               'latitude': 47.2755434,
                                               'longitude': -2.2022817},
                                       'name': 'VIP',
                                       'sameAs': 'http://www.levip-saintnazaire.com/'},
                          'name': 'Elysian Fields',
                          'performer': [{'@type': 'MusicGroup',
                                         'name': 'Elysian Fields',
                                         'sameAs': 'https://www.songkick.com/artists/236156-elysian-fields?utm_medium=organic&utm_source=microformat'},
                                        {'@type': 'MusicGroup',
                                         'name': 'Troy Von Balthazar',
                                         'sameAs': 'https://www.songkick.com/artists/355304-troy-von-balthazar?utm_medium=organic&utm_source=microformat'}],
                          'startDate': '2016-10-29T21:00:00+0200',
                          'url': 'https://www.songkick.com/concerts/27626524-elysian-fields-at-vip?utm_medium=organic&utm_source=microformat'},
                         {'@context': 'http://schema.org',
                          '@type': 'MusicGroup',
                          'image': 'https://images.sk-static.com/images/media/profile_images/artists/236156/card_avatar',
                          'interactionCount': '6066 UserLikes',
                          'logo': 'https://images.sk-static.com/images/media/profile_images/artists/236156/card_avatar',
                          'name': 'Elysian Fields',
                          'url': 'https://www.songkick.com/artists/236156-elysian-fields?utm_medium=organic&utm_source=microformat'}]
        self.assertEqual(json_ld_data, expected_ld_data)

    # value of microdata extraction is a list
    def test_rdfa_extrator(self) -> None:
        hme = HTMLMetadataExtractor()
        test_html = """<html>
                    ...  <head>
                    ...    ...
                    ...  </head>
                    ...  <body prefix="dc: http://purl.org/dc/terms/ schema: http://schema.org/">
                    ...    <div resource="/alice/posts/trouble_with_bob" typeof="schema:BlogPosting">
                    ...       <h2 property="dc:title">The trouble with Bob</h2>
                    ...       <h3 property="dc:creator schema:creator" resource="#me">Alice</h3>
                    ...       <div property="schema:articleBody">
                    ...         <p>The trouble with Bob is that he takes much better photos than I do:</p>
                    ...       </div>
                    ...    </div>
                    ...  </body>
                    ... </html>
                    ... """

        rdfa_data = hme.extract(test_html, extract_rdfa = True)[0].value
        expected_rdfa_data = [{'@id': 'http://www.example.com/alice/posts/trouble_with_bob',
                                        '@type': ['http://schema.org/BlogPosting'], 
                                        'http://purl.org/dc/terms/creator': [{'@id': 'http://www.example.com/#me'}], 
                                        'http://purl.org/dc/terms/title': [{'@value': 'The trouble with Bob'}], 
                                        'http://schema.org/articleBody': [{'@value': '\n                    ...         The trouble with Bob is that he takes much better photos than I do:\n                    ...       '}],
                                        'http://schema.org/creator': [{'@id': 'http://www.example.com/#me'}]}]
        self.assertEqual(rdfa_data, expected_rdfa_data)


if __name__ == '__main__':
    unittest.main()


import re

from etk.extractors.regex_extractor import RegexExtractor


class CVEExtractor(RegexExtractor):
    """
    **Description**
        This class inherits RegexExtractor by predefine the CVE regex pattern
        CVE id form: https://cve.mitre.org/about/faqs.html#what_is_cve_id

    Examples:
        ::

            CVE_extractor = CVEExtractor()
            CVE_extractor.extract(text=input_doc)

    """
    def __init__(self):
        cve_pattern = r"CVE-(?:\d{4})-(?:\d{4,7})"
        RegexExtractor.__init__(self, pattern=cve_pattern, flags=re.IGNORECASE,
                                extractor_name="cve extractor")



from typing import List, Dict
import numbers, re
from datetime import date, datetime
from etk.field_types import FieldType
from etk.etk_exceptions import ISODateError


class KGSchema(object):
    """
    This class define the schema for a knowledge graph object.
    Create a knowledge graph schema according to the master config the user defined in myDIG UI
    """

    def __init__(self, config: Dict) -> None:
        """
        Record a mapping about each fields and its type from config file

        Args:
            config: Dict
        """

        self.fields_dict = dict()
        try:
            for field in config["fields"]:
                if config["fields"][field]["type"] == "kg_id":
                    self.fields_dict[field] = FieldType.KG_ID
                elif config["fields"][field]["type"] == "number":
                    self.fields_dict[field] = FieldType.NUMBER
                elif config["fields"][field]["type"] == "date":
                    self.fields_dict[field] = FieldType.DATE
                elif config["fields"][field]["type"] == "location":
                    self.fields_dict[field] = FieldType.LOCATION
                else:
                    self.fields_dict[field] = FieldType.STRING

        except KeyError as key:
            print(str(key) + " not in config")

    @property
    def fields(self) -> List[str]:
        """
        Return a list of all fields that are defined in master config

        Args:


        Returns: List of fields
        """
        return list(self.fields_dict.keys())

    def has_field(self, field_name: str) -> bool:
        """
        Return true if the schema has the field, otherwise false

        Args:
            field_name: str

        Returns: bool
        """
        return field_name in self.fields_dict

    def field_type(self, field_name: str) -> FieldType:
        """
        Return the type of a field defined in schema, if field not defined, return None

        Args:
            field_name: str

        Returns: FieldType
        """
        return self.fields_dict.get(field_name, None)

        # Amandeep: The code below does not do what the description of the function says, replaced with line above
        # if self.has_field(field_name):
        #     return self.fields_dict[field_name]
        # else:
        #     print(field_name + " field not defined")

    @staticmethod
    def iso_date(d) -> str:
        """
        Return iso format of a date

        Args:
            d:
        Returns: str

        """
        if isinstance(d, datetime):
            return d.isoformat()
        elif isinstance(d, date):
            return datetime.combine(d, datetime.min.time()).isoformat()
        else:
            try:
                datetime.strptime(d, '%Y-%m-%dT%H:%M:%S')
                return d
            except ValueError:
                try:
                    datetime.strptime(d, '%Y-%m-%d')
                    return d + "T00:00:00"
                except ValueError:
                    pass
        raise ISODateError("Can not convert value to ISO format for kg")

    @staticmethod
    def parse_number(txt_num):
        if isinstance(txt_num, str) or isinstance(txt_num, numbers.Number):

            if isinstance(txt_num, numbers.Number):
                return str(txt_num)

            try:
                # this will fail spectacularly if the data originated in europe,
                # where they use ',' inplace of '.' and vice versa for numbers

                text = txt_num.strip().replace('\n', '').replace('\t', '').replace(',', '')
                num = str(float(text)) if '.' in text else str(int(text))
                return num
            except:
                pass
        return None

    def is_valid(self, field_name, value) -> (bool, object):
        """
        Return true if the value type matches or can be coerced to the defined type in schema, otherwise false.
        If field not defined, return none

        Args:
            field_name: str
            value:

        Returns: bool, value, where the value may have been coerced to the required type.
        """

        if self.has_field(field_name):
            if self.fields_dict[field_name] == FieldType.KG_ID:
                return True, value

            if self.fields_dict[field_name] == FieldType.NUMBER:
                if isinstance(value, numbers.Number):
                    return True, value
                else:
                    converted_number = self.parse_number(value)
                    return (False, value) if not converted_number else (True, value)
            if self.fields_dict[field_name] == FieldType.STRING:
                if isinstance(value, str):
                    return True, value.strip()
                else:
                    return True, str(value).strip()

            if self.fields_dict[field_name] == FieldType.DATE:
                valid, d = self.is_date(value)
                if valid:
                    return True, d.isoformat()
                else:
                    return False, value

            if self.fields_dict[field_name] == FieldType.LOCATION:
                valid, l = self.is_location(value)
                if valid:
                    return True, l
                else:
                    return False, value
        else:
            print('{} not found in KG Schema'.format(field_name))
            return False, value
            # Amandeep: returning False instead of printing a message
            # print(field_name + " field not defined")

    @staticmethod
    def is_date(v) -> (bool, date):
        """
        Boolean function for checking if v is a date

        Args:
            v:
        Returns: bool

        """
        if isinstance(v, date):
            return True, v
        try:
            reg = r'^([0-9]{4})(?:-(0[1-9]|1[0-2])(?:-(0[1-9]|[1-2][0-9]|3[0-1])(?:T' \
                  r'([0-5][0-9])(?::([0-5][0-9])(?::([0-5][0-9]))?)?)?)?)?$'
            match = re.match(reg, v)
            if match:
                groups = match.groups()
                patterns = ['%Y', '%m', '%d', '%H', '%M', '%S']
                d = datetime.strptime('-'.join([x for x in groups if x]),
                                      '-'.join([patterns[i] for i in range(len(patterns)) if groups[i]]))
                return True, d
        except:
            pass
        return False, v

    @staticmethod
    def is_location(v) -> (bool, str):
        """
        Boolean function for checking if v is a location format

        Args:
            v:
        Returns: bool

        """

        def convert2float(value):
            try:
                float_num = float(value)
                return float_num
            except ValueError:
                return False

        if not isinstance(v, str):
            return False, v
        split_lst = v.split(":")
        if len(split_lst) != 5:
            return False, v
        if convert2float(split_lst[3]):
            longitude = abs(convert2float(split_lst[3]))
            if longitude > 90:
                return False, v
        if convert2float(split_lst[4]):
            latitude = abs(convert2float(split_lst[3]))
            if latitude > 180:
                return False, v
        return True, v



from etk.extraction import Extractable
from typing import List, Dict

from etk.origin_record import OriginRecord


class ExtractionProvenanceRecord(Extractable):
    """
    A Provenance Record containing details of Extraction Results history.
    """

    def __init__(self, id: int, json_path: str, method: str, start_char: str, end_char: str, confidence, _document,
                 parent_extraction_provenance: int = None) -> None:

        Extractable.__init__(self)
        self.id = id
        self.origin_record = OriginRecord(json_path, start_char, end_char, _document)
        self.method = method
        self.extraction_confidence = confidence
        self._document = _document
        self._parent_extraction_provenance = parent_extraction_provenance

    @property
    def parent_extraction_provenance(self):
        """
        Returns: the parent extraction provenance id.
        """
        return self._parent_extraction_provenance

    def get_origins(self, value):
        if self._parent_extraction_provenance is None:
            return [self.origin_record]
        else:
            parent_id = self._parent_extraction_provenance
            parent = self._document.provenances[parent_id]
            origins = parent.get_origins(value)
            return origins



import unittest
from etk.extractors.spacy_rule_extractor import SpacyRuleExtractor
import spacy
import json

rules = json.load(open('etk/unit_tests/ground_truth/spacy_rules.json', "r"))


class TestSpacyRuleExtractor(unittest.TestCase):

    def setUp(self):
        self.nlp = spacy.load("en_core_web_sm")

    def test_SpacyRuleExtractor_word_1(self) -> None:
        sample_rules = rules["test_SpacyRuleExtractor_word_1"]
        sample_rule_extractor = SpacyRuleExtractor(self.nlp, sample_rules, "test_extractor")
        extractions = sample_rule_extractor.extract(
            "version 2 of etk, implemented by Runqi Shao, Dongyu Li, Sylvia lin, Amandeep and others.")

        expected = [('rule_0', 'Name: Runqi Shao'), ('rule_0', 'Name: Dongyu Li'), ('rule_0', 'Name: Sylvia'),
                    ('rule_0', 'Name: Amandeep')]
        self.assertEqual([(x.rule_id, x.value) for x in extractions], expected)

    def test_SpacyRuleExtractor_word_2(self) -> None:
        sample_rules = rules["test_SpacyRuleExtractor_word_2"]
        sample_rule_extractor = SpacyRuleExtractor(self.nlp, sample_rules, "test_extractor")
        extractions = sample_rule_extractor.extract(
            "version 2 of etk, implemented by Runqi Shao, Dongyu Li, Sylvia lin, Amandeep and others.")
        expected = [('rule_0', 'Name: Sylvia lin')]
        self.assertEqual([(x.rule_id, x.value) for x in extractions], expected)

    def test_SpacyRuleExtractor_word_3(self) -> None:
        sample_rules = rules["test_SpacyRuleExtractor_word_3"]

        sample_rule_extractor = SpacyRuleExtractor(self.nlp, sample_rules, "test_extractor")
        extractions = sample_rule_extractor.extract(
            "version 2 of etk, implemented by Runqi Shao, DongYu94 Li, Sylvia lin, Amandeep and others.")

        expected = [('rule_1', 'Name: Runqi Shao'), ('rule_0', 'Name: DongYu94 Li')]
        self.assertEqual([(x.rule_id, x.value) for x in extractions], expected)

    def test_SpacyRuleExtractor_word_4(self) -> None:
        sample_rules = rules["test_SpacyRuleExtractor_word_4"]

        sample_rule_extractor = SpacyRuleExtractor(self.nlp, sample_rules, "test_extractor")
        extractions = sample_rule_extractor.extract(
            "version 2 of etk, implemented by Runqi Shao, DongYu94 Li, Sylvia lin, Amandeep and others.")

        expected = [('rule_0', 'Name: Sylvia'), ('rule_0', 'Name: Amandeep')]
        self.assertEqual([(x.rule_id, x.value) for x in extractions], expected)

    def test_SpacyRuleExtractor_word_5(self) -> None:
        sample_rules = rules["test_SpacyRuleExtractor_word_5"]

        sample_rule_extractor = SpacyRuleExtractor(self.nlp, sample_rules, "test_extractor")
        extractions = sample_rule_extractor.extract(
            "version 2 of etk, implemented by Runqi Shao, DongYu94 Li, Sylvia lin, Amandeep and others.")
        expected = [('rule_0', 'First Name: Runqi, Last Name: Shao. Full name: Runqi Shao'),
                    ('rule_0', 'First Name: DongYu94, Last Name: Li. Full name: DongYu94 Li')]
        self.assertEqual([(x.rule_id, x.value) for x in extractions], expected)

    def test_SpacyRuleExtractor_number_1(self) -> None:
        sample_rules = rules["test_SpacyRuleExtractor_number_1"]

        sample_rule_extractor = SpacyRuleExtractor(self.nlp, sample_rules, "test_extractor")
        extractions = sample_rule_extractor.extract(
            "Extract from the following number: 32 12 54435 23 665.3 34 65.42 23 4545")

        expected = [('rule_0', '665.3'), ('rule_0', '4545')]
        self.assertEqual([(x.rule_id, x.value) for x in extractions], expected)

    def test_SpacyRuleExtractor_number_2(self) -> None:
        sample_rules = rules["test_SpacyRuleExtractor_number_2"]

        sample_rule_extractor = SpacyRuleExtractor(self.nlp, sample_rules, "test_extractor")
        extractions = sample_rule_extractor.extract(
            "Extract from the following number: 32 12 54435 23 665.3 34 65.42 23 4545")

        expected = [('rule_0', '32'), ('rule_0', '12'), ('rule_0', '23'), ('rule_0', '34'), ('rule_0', '23'), ('rule_0', '4545')]
        self.assertEqual([(x.rule_id, x.value) for x in extractions], expected)

    def test_SpacyRuleExtractor_shape_1(self) -> None:
        sample_rules = rules["test_SpacyRuleExtractor_shape_1"]

        sample_rule_extractor = SpacyRuleExtractor(self.nlp, sample_rules, "test_extractor")
        extractions = sample_rule_extractor.extract(
            "version 2 of etk, implemented by RqS, DongYu94 Li, Sylvia lin, Amandeep and others.")

        expected = [('rule_0', 'RqS'), ('rule_0', 'DongYu94')]
        self.assertEqual([(x.rule_id, x.value) for x in extractions], expected)

    def test_SpacyRuleExtractor_punc_1(self) -> None:
        sample_rules = rules["test_SpacyRuleExtractor_punc_1"]

        sample_rule_extractor = SpacyRuleExtractor(self.nlp, sample_rules, "test_extractor")
        extractions = sample_rule_extractor.extract(
            "version 2 of etk, implemented by Rq? Shao. DongYu94 Li, Sylvia-lin, Amandeep and others.")

        expected = [('rule_0', 'Name: Rq, Shao'), ('rule_0', 'Name: Sylvia, lin')]
        self.assertEqual([(x.rule_id, x.value) for x in extractions], expected)

    def test_SpacyRuleExtractor_linebreak_1(self) -> None:
        sample_rules = rules["test_SpacyRuleExtractor_linebreak_1"]

        sample_rule_extractor = SpacyRuleExtractor(self.nlp, sample_rules, "test_extractor")
        extractions = sample_rule_extractor.extract(
            "version 2 of etk, implemented by Rq? Shao. DongYu94 Li, \n\n\n Sylvia-lin, Amandeep and others.")

        expected = [('rule_0', 'Length 3 linebreak: \n\n\n ')]
        self.assertEqual([(x.rule_id, x.value) for x in extractions], expected)

    def test_SpacyRuleExtractor_output_format_1(self) -> None:
        sample_rules = rules["test_SpacyRuleExtractor_output_format_1"]

        sample_rule_extractor = SpacyRuleExtractor(self.nlp, sample_rules, "test_extractor")
        extractions = sample_rule_extractor.extract(
            "version 2 of etk, implemented by Runqi Shao, Dongyu Li, Sylvia lin, Amandeep and others.")

        expected = [('rule_0', 'Name: Sylvia, lin')]
        self.assertEqual([(x.rule_id, x.value) for x in extractions], expected)

    def test_SpacyRuleExtractor_output_format_2(self) -> None:
        sample_rules = rules["test_SpacyRuleExtractor_output_format_2"]

        sample_rule_extractor = SpacyRuleExtractor(self.nlp, sample_rules, "test_extractor")
        extractions = sample_rule_extractor.extract(
            "version 2 of etk, implemented by Runqi Shao, Dongyu Li, Sylvia lin, Amandeep and others.")

        expected = [('rule_0', 'Name: Sylvia, lin')]
        self.assertEqual([(x.rule_id, x.value) for x in extractions], expected)

    def test_SpacyRuleExtractor_output_format_3(self) -> None:
        sample_rules = rules["test_SpacyRuleExtractor_output_format_3"]

        sample_rule_extractor = SpacyRuleExtractor(self.nlp, sample_rules, "test_extractor")
        extractions = sample_rule_extractor.extract(
            "version 2 of etk, implemented by Runqi Shao, Dongyu Li, Sylvia lin, Amandeep and others.")

        expected = [('rule_0', 'Name: Sylvia, , lin')]
        self.assertEqual([(x.rule_id, x.value) for x in extractions], expected)


if __name__ == '__main__':
    unittest.main()



singleton_regex = {
    '%Y': r'([1-2][0-9][0-9][0-9])',  # year in four digits
    '%y': r'([6-9][0-9]|[0-3][0-9])',  # year in two digits
    '%B': r'(January|February|March|April|May|June|July|August|September|October|November|December|'    # month
          r'enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre|'  # spainish
          r'januari|februari|mars|april|maj|juni|juli|augusti|september|oktober|november|december)',    # swedish
    '%b': r'(Jan\.?|Feb\.?|Mar\.?|Apr\.?|Jun\.?|Jul\.?|Aug\.?|Sep(?:t?)\.?|Oct\.?|Nov\.?|Dec\.?|'
          r'enero|feb\.?|marzo|abr\.?|mayo|jun\.?|jul\.?|agosto|sept\.?|set\.?|oct\.?|nov\.?|dic\.?)',  # month abbr.
    '%m': r'(1[0-2]|0?[1-9])',  # month in two digits: 01-12 or just 0-12
    '%d': r'(3[0-1]|[1-2][0-9]|0?[1-9])',  # day in two digits: 01-31 or juest 1-31
    'd_suffix': r'(?:st|nd|rd|th)',
    '%A': r'(Monday|Tuesday|Wednesday|Thursday|Friday|Saturday|Sunday|'
          r'lunes|martes|miércoles|jueves|viernes|sábado|domingo|'
          r'Måndag|Tisdag|Onsdag|Torsdag|Fredag|Lördag|Söndag)',  # weekdays
    '%a': r'(Mon\.?|Tue\.?|Wed\.?|Thu(?:r(?:s?)?)?\.?|Fri\.?|Sat\.?|Sun\.?|L|M|X|J|V|S|D)',  # weekdays abbr.
    '%H': r'(2[0-3]|1[0-9]|0?[0-9])',  # hour in 24-hours in two digits: 00-23 or just 0-23
    '%I': r'(1[0-2]|0?[1-9])',  # hour in 12-hour in two digits: 01-12 or just 1-12
    '%p': r'((?: ?)AM(?:\.?)|(?: ?)PM(?:\.?))',  # am/pm markers
    '%M': r'([0-5][0-9])',  # minute in two digits: 00-59
    '%S': r'([0-5][0-9](?:\.[0-9]+)?)',  # second in two digits: 00-59 or with fractions like 12.355
    '%Z': r'(ACDT|ACST|ACT|ACT|ACWST|ADT|AEDT|AEST|AFT|AKDT|AKST|AMST|AMT|AMT|ART|AST|AST|AWST|AZOST|AZOT|AZT|'
          r'BDT|BIOT|BIT|BOT|BRST|BRT|BST|BST|BST|BTT|CAT|CCT|CDT|CDT|CEST|CET|CHADT|CHAST|CHOT|CHOST|CHST|CHUT|'
          r'CIST|CIT|CKT|CLST|CLT|COST|COT|CST|CST|CST|CT|CVT|CWST|CXT|DAVT|DDUT|DFT|'
          r'EASST|EAST|EAT|ECT|ECT|EDT|EEST|EET|EGST|EGT|EIT|EST|FET|FJT|FKST|FKT|FNT|'
          r'GALT|GAMT|GET|GFT|GILT|GIT|GMT|GST|GST|GYT|HDT|HAEC|HST|HKT|HMT|HOVST|HOVT|'
          r'ICT|IDT|IOT|IRDT|IRKT|IRST|IST|IST|IST|JST|KGT|KOST|KRAT|KST|LHST|LHST|LINT|'
          r'MAGT|MART|MAWT|MDT|MET|MEST|MHT|MIST|MIT|MMT|MSK|MST|MST|MUT|MVT|MYT|NCT|NDT|NFT|NPT|NST|NT|NUT|NZDT|NZST|'
          r'OMST|ORAT|PDT|PET|PETT|PGT|PHOT|PHT|PKT|PMDT|PMST|PONT|PST|PST|PYST|PYT|RET|ROTT|'
          r'SAKT|SAMT|SAST|SBT|SCT|SDT|SGT|SLST|SRET|SRT|SST|SST|SYOT|TAHT|THA|TFT|TJT|TKT|TLT|TMT|TRT|TOT|TVT|'
          r'ULAST|ULAT|USZ1|UTC|UYST|UYT|UZT|VET|VLAT|VOLT|VOST|VUT|WAKT|WAST|WAT|WEST|WET|WIT|WST|YAKT|YEKT)',
    # timezone abbr. list
    # reference: https://en.wikipedia.org/wiki/List_of_time_zone_abbreviations
    '%z': r'((?:UTC|GMT)(?: ?[\+\-] ?(?:(?:1[0-4]|0?[0-9])(?::?(?:00|30|45))?))?|'
          r'[+-][01][0-3](?:00|30|45)|[\+\-](?:1[0-3]|0[0-9])(?:00|30|45))',
    # timezone like 'UTC', 'UTC + 8:30', 'GMT+1130', '+0600'
    # reference: https://en.wikipedia.org/wiki/Time_zone
    # 'splitters': r'(?:[,/ \.\-]? ?)',
    'splitters': r'(?:,? of |,? del |,? de |,? el |\. |, |-|/| ?)',
    'y_prefix': r'(?:in |at |by |year )',
    'w_prefix': r'(?:on )',
}

# TODO: reconstruct the regex generater to filter out the false positives on first time matching, to speed up
illegal = '^\b?(?:[a-zA-Z]?[0-9]{1,3}[a-zA-Z]?|[0-9]{5}|[0-9]{7})\b?$'
possible_illegal = '^\b?(?:[0-9]{1,6}[-/ ,\.] ?[0-9]{1,6})\b?$'
possible_illegal_3 = '^\b?(?:[0-9]{1,2}[-/ ,\.] ?[0-9]{1,2}[-/ ,\.] ?[0-9]{1,2})\b?$'

units = {
    'Y': ['%Y', '%y'],
    'M': ['%B', '%b', '%m'],
    'D': ['%d', ['%d', 'd_suffix']],
    'W': ['%A', '%a'],
    'HOUR': ['%I', '%H'],
    'MIN': '%M',
    'SEC': '%S',
    'MARK': '%p',
    'TZ': ['%Z', '%z'],
    'SINGLE_YEAR': [['y_prefix', '%Y']],
    'SINGLE_WEEK': ['%A', ['w_prefix', '%a']],
    'SINGLE_MONTH': ['%B', '%b']
}

day_of_week_to_number = {
    'monday': 0,
    'tuesday': 1,
    'wednesday': 2,
    'thursday': 3,
    'friday': 4,
    'saturday': 5,
    'sunday': 6
}

foreign_to_english = {
    # spanish
    'enero': 'January',
    'febrero': 'February',
    'marzo': 'March',
    'abril': 'April',
    'mayo': 'May',
    'junio': 'June',
    'julio': 'July',
    'agosto': 'August',
    'septiembre': 'September',
    'octubre': 'October',
    'noviembre': 'November',
    'diciembre': 'December',
    # 'feb': 'Feb',
    'abr': 'Apr',
    'set': 'Sept',
    'dic': 'Dec',
    'lunes': 'Monday',
    'martes': 'Tuesday',
    'miércoles': 'Wednesday',
    'jueves': 'Thursday',
    'viernes': 'Friday',
    'sábado': 'Saturday',
    'domingo': 'Sunday',
    'L': 'Mon',
    'M': 'Tue',
    'X': 'Wed',
    'J': 'Thu',
    'V': 'Fri',
    'S': 'Sat',
    'D': 'Sun',

    # swedish
    'januari': 'January',
    'februari': 'February',
    'mars': 'March',
    # 'april': 'April',
    'maj': 'May',
    'juni': 'June',
    'juli': 'July',
    'augusti': 'August',
    # 'september': 'September',
    'oktober': 'October',
    # 'november': 'November',
    # 'december': 'December',
    'måndag': 'Monday',
    'tisdag': 'Tuesday',
    'onsdag': 'Wednesday',
    'torsdag': 'Thursday',
    'fredag': 'Friday',
    'lördag': 'Saturday',
    'söndag': 'Sunday'
}

language_date_order = {
    'es': 'DMY',
    'en': 'MDY'
}

# relative date:
# before, after, last, next, ago, later, in
# number, half, quarter, several?, some?,
# weeks, years, days, months, hours, minutes, seconds
# now, , yesterday, tomorrow
directions = {
    'last': '-',
    'next': '+',
    'before': '-',
    'after': '+',
    'in': '+',
    'ago': '-',
    'later': '+'
}
num_to_digit = {
    'a': '1',
    'an': '1',
    'one': '1',
    'two': '2',
    'three': '3',
    'four': '4',
    'five': '5',
    'six': '6',
    'seven': '7',
    'eight': '8',
    'nine': '9',
    'ten': '10'
}
spacy_rules = {
  "field_name": "relative_date",
  "rules": [
    {
      "dependencies": [],
      "description": "",
      "identifier": "direction_number_unit",
      "is_active": "true",
      "output_format": "",
      "pattern": [
        {
          "capitalization": [],
          "contain_digit": "false",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "true",
          "length": [],
          "match_all_forms": "true",
          "maximum": "",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": ["in", "after", "before"
          ],
          "type": "word"
        },
        {
          "capitalization": [],
          "contain_digit": "false",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "true",
          "length": [],
          "match_all_forms": "true",
          "maximum": "",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": ["a", "an", "one", "two", "three", "four", "five", "six", "seven", "eight", "nine", "ten"],
          "type": "word"
        },
        {
          "capitalization": [],
          "contain_digit": "false",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "true",
          "length": [],
          "match_all_forms": "true",
          "maximum": "",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": ["weeks", "years", "days", "months", "hours", "minutes", "seconds",
                    "week", "year", "day", "month", "hour", "minute", "second"],
          "type": "word"
        }
      ],
      "polarity": "true"
    },
    {
      "dependencies": [],
      "description": "",
      "identifier": "number_unit_direction",
      "is_active": "true",
      "output_format": "",
      "pattern": [
        {
          "capitalization": [],
          "contain_digit": "false",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "true",
          "length": [],
          "match_all_forms": "true",
          "maximum": "",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": ["a", "an", "one", "two", "three", "four", "five", "six", "seven", "eight", "nine", "ten"],
          "type": "word"
        },
        {
          "capitalization": [],
          "contain_digit": "false",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "true",
          "length": [],
          "match_all_forms": "true",
          "maximum": "",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": ["weeks", "years", "days", "months", "hours", "minutes", "seconds",
                    "week", "year", "day", "month", "hour", "minute", "second"],
          "type": "word"
        },
        {
          "capitalization": [],
          "contain_digit": "false",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "true",
          "length": [],
          "match_all_forms": "true",
          "maximum": "",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": ["before", "ago", "later", "after"],
          "type": "word"
        }
      ],
      "polarity": "true"
    },
    {
      "dependencies": [],
      "description": "",
      "identifier": "direction_digit_unit",
      "is_active": "true",
      "output_format": "",
      "pattern": [
        {
          "capitalization": [],
          "contain_digit": "false",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "true",
          "length": [],
          "match_all_forms": "true",
          "maximum": "",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": ["in", "after", "before"
          ],
          "type": "word"
        },
        {
          "capitalization": [],
          "contain_digit": "true",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "true",
          "length": [],
          "match_all_forms": "true",
          "maximum": "0",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": [
          ],
          "type": "number"
        },
        {
          "capitalization": [],
          "contain_digit": "false",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "true",
          "length": [],
          "match_all_forms": "true",
          "maximum": "",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": ["weeks", "years", "days", "months", "hours", "minutes", "seconds",
                    "week", "year", "day", "month", "hour", "minute", "second"],
          "type": "word"
        }
      ],
      "polarity": "true"
    },
    {
      "dependencies": [],
      "description": "",
      "identifier": "digit_unit_direction",
      "is_active": "true",
      "output_format": "",
      "pattern": [
        {
          "capitalization": [],
          "contain_digit": "true",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "true",
          "length": [],
          "match_all_forms": "true",
          "maximum": "0",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": [
          ],
          "type": "number"
        },
        {
          "capitalization": [],
          "contain_digit": "false",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "true",
          "length": [],
          "match_all_forms": "true",
          "maximum": "",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": ["weeks", "years", "days", "months", "hours", "minutes", "seconds",
                    "week", "year", "day", "month", "hour", "minute", "second"],
          "type": "word"
        },
        {
          "capitalization": [],
          "contain_digit": "false",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "true",
          "length": [],
          "match_all_forms": "true",
          "maximum": "",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": ["before", "ago", "later", "after"],
          "type": "word"
        }
      ],
      "polarity": "true"
    },
    {
      "dependencies": [],
      "description": "",
      "identifier": "direction_unit",
      "is_active": "true",
      "output_format": "",
      "pattern": [
        {
          "capitalization": [],
          "contain_digit": "false",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "true",
          "length": [],
          "match_all_forms": "true",
          "maximum": "",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": ["last", "next"
          ],
          "type": "word"
        },
        {
          "capitalization": [],
          "contain_digit": "false",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "true",
          "length": [],
          "match_all_forms": "true",
          "maximum": "",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": ["week", "year", "day", "month", "hour", "minute", "second"],
          "type": "word"
        }
      ],
      "polarity": "true"
    },
    {
      "dependencies": [],
      "description": "",
      "identifier": "the_day",
      "is_active": "true",
      "output_format": "",
      "pattern": [
    {
          "capitalization": [],
          "contain_digit": "false",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "false",
          "length": [],
          "match_all_forms": "true",
          "maximum": "",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": ["the"],
          "type": "word"
        },
        {
          "capitalization": [],
          "contain_digit": "false",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "false",
          "length": [],
          "match_all_forms": "true",
          "maximum": "",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": ["day"],
          "type": "word"
        },
        {
          "capitalization": [],
          "contain_digit": "false",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "false",
          "length": [],
          "match_all_forms": "true",
          "maximum": "",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": ["before", "after"],
          "type": "word"
        },
        {
          "capitalization": [],
          "contain_digit": "false",
          "is_in_output": "true",
          "is_in_vocabulary": "false",
          "is_out_of_vocabulary": "false",
          "is_required": "true",
          "length": [],
          "match_all_forms": "true",
          "maximum": "",
          "minimum": "",
          "numbers": [],
          "part_of_speech": [],
          "prefix": "",
          "shapes": [
          ],
          "suffix": "",
          "token": ["yesterday", "today", "tomorrow"],
          "type": "word"
        },
      ],
      "polarity": "true"
    }
  ]
}


import unittest
from etk.extractors.spacy_ner_extractor import SpacyNerExtractor


class TestSpacyNerExtractor(unittest.TestCase):

    def test_spacy_ner_extractor(self) -> None:
        get_attr = ['PERSON', 'ORG', 'GPE']
        extractor = SpacyNerExtractor(extractor_name='spacy_ner_extractor')
        text = 'Napoléon Bonaparte was a French statesman and military leader who rose to prominence during the French Revolution and led several successful campaigns during the French Revolutionary Wars. As Napoleon, he was Emperor of the French from 1804 until 1814, and again briefly in 1815 during the Hundred Days. Napoleon dominated European and global affairs for more than a decade while leading France against a series of coalitions in the Napoleonic Wars. He won most of these wars and the vast majority of his battles, building a large empire that ruled over continental Europe before its final collapse in 1815. He is considered one of the greatest commanders in history, and his wars and campaigns are studied at military schools worldwide. Napoleon\'s political and cultural legacy has endured as one of the most celebrated and controversial leaders in human history.'
        extracted = list()
        results = extractor.extract(text, get_attr=get_attr)

        for i in results:
            extracted_value = {
                'value': i.value,
                'start_char': i.provenance['start_char'],
                'end_char': i.provenance['end_char'],
                'start_token': i.provenance['start_token'],
                'end_token': i.provenance['end_token'],
                'tag': i.tag
            }
            extracted.append(extracted_value)
        expected = [{'value': 'Napoléon Bonaparte', 'start_char': 0, 'end_char': 18, 'start_token': 0, 'end_token': 2,
                     'tag': 'PERSON'},
                    {'value': 'Napoleon', 'start_char': 192, 'end_char': 200, 'start_token': 29, 'end_token': 30,
                     'tag': 'ORG'},
                    {'value': 'Napoleon', 'start_char': 304, 'end_char': 312, 'start_token': 52, 'end_token': 53,
                     'tag': 'ORG'},
                    {'value': 'France', 'start_char': 388, 'end_char': 394, 'start_token': 65, 'end_token': 66,
                     'tag': 'GPE'},
                    {'value': 'Napoleon', 'start_char': 738, 'end_char': 746, 'start_token': 129, 'end_token': 130,
                     'tag': 'ORG'}]
        result_count = 0
        while result_count < len(extracted):
            self.assertEqual(extracted[result_count], expected[result_count])
            result_count += 1


if __name__ == '__main__':
    unittest.main()



from etk.dependencies.date_extractor_resources.constants import singleton_regex, units


class DateRegexGenerator:
    """
    get the final_regex and symbol_list for date_extractor if single regex updates
    d = DateRegexGenerator(singleton_regex, units)
    final_regex = d.final_regex
    symbol_list = d.single_list
    """
    def __init__(self, singleton_regex, units):
        self.singleton_regex = singleton_regex
        self.units = units

        temp = self.generate_all_regexes_by_singletons()
        self.final_regex = temp[0]
        self.symbol_list = temp[1]

    def generate_all_regexes_by_singletons(self):
        all = {}
        for k in self.units.keys():
            unit = self.generate_regex_for_a_unit(self.units[k])
            all[k] = unit
        s = self.singleton_regex['splitters']

        time = r'(?:' + all['HOUR']['res'] + r': ?' + all['MIN']['res'] + r'(?:: ?' + all['SEC']['res'] + r')?(?:' + \
               all['MARK']['res'] + ')?(?:' + s + all['TZ']['res'] + r')?)'
        st = r'(?:' + self.singleton_regex['splitters'] + r'|T)?'
        time_reg = r'(?:T?' + st + time + r'Z?)?'

        week_reg_post = r'(?:' + s + all['W']['res'] + r')?'
        week_reg_pre = r'(?:' + all['W']['res'] + s + r')?'

        day_reg_post = r'(?:' + s + all['D']['res'] + r')?'
        day_reg_pre = r'(?:' + all['D']['res'] + s + r')?'

        year_reg_post = r'(?:' + s + all['Y']['res'] + r')?'
        year_reg_pre = r'(?:' + all['Y']['res'] + s + r')?'

        pl = 'pattern_list'
        time_pattern = all['HOUR'][pl] + all['MIN'][pl] + all['SEC'][pl] + all['MARK'][pl] + all['TZ'][pl]

        return ({
                    'MDY': r'(?<=\b)(?:(?:' + week_reg_pre + all['M']['res'] + day_reg_post +
                           year_reg_post + week_reg_post + r')' + time_reg + r')(?=\b)',
                    'DMY': r'(?<=\b)(?:(?:' + week_reg_pre + day_reg_pre + all['M']['res'] +
                           year_reg_post + week_reg_post + r')' + time_reg + r')(?=\b)',
                    'YMD': r'(?<=\b)(?:(?:' + week_reg_pre + year_reg_pre + all['M'][
                        'res'] + day_reg_post + week_reg_post + r')' + time_reg + r')(?=\b)',
                    'SINGLE_YEAR': r'(?<=\b)(?:(?:' + all['SINGLE_YEAR']['res'] + r')' + time_reg + r')(?=\b)',
                    'SINGLE_MONTH': r'(?<=\b)(?:(?:' + all['SINGLE_MONTH']['res'] + r')' + time_reg + r')(?=\b)',
                    'SINGLE_WEEK': r'(?<=\b)(?:(?:' + all['SINGLE_WEEK']['res'] + r')' + time_reg + r')(?=\b)'
                }, {
                    'MDY': all['W'][pl] + all['M'][pl] + all['D'][pl] + all['Y'][pl] + all['W'][pl] + time_pattern,
                    'DMY': all['W'][pl] + all['D'][pl] + all['M'][pl] + all['Y'][pl] + all['W'][pl] + time_pattern,
                    'YMD': all['W'][pl] + all['Y'][pl] + all['M'][pl] + all['D'][pl] + all['W'][pl] + time_pattern,
                    'SINGLE_YEAR': all['SINGLE_YEAR'][pl] + time_pattern,
                    'SINGLE_MONTH': all['SINGLE_MONTH'][pl] + time_pattern,
                    'SINGLE_WEEK': all['SINGLE_WEEK'][pl] + time_pattern
                })

    def generate_regex_for_a_unit(self, key_list):
        if isinstance(key_list, list):
            if not key_list or len(key_list) == 0:
                return {'res': r'', 'pattern_list': []}
            pattern_list = []
            regex_list = []
            for key in key_list:
                if isinstance(key, list):
                    regex = self.singleton_regex[key[0]]
                    if key[0][0] == '%':
                        pattern_list.append(key[0])
                    for sub_key in key[1:]:
                        regex = regex + self.singleton_regex[sub_key]
                        if sub_key[0] == '%':
                            pattern_list.append(sub_key)
                    regex = regex
                    regex_list.append(regex)
                else:
                    regex_list.append(self.singleton_regex[key])
                    if key[0] == '%':
                        pattern_list.append(key)
            if not regex_list or len(regex_list) == 0:
                return {'res': r'', 'pattern_list': []}
            res = r'(?:' + regex_list[0]
            for r in regex_list[1:]:
                res = res + '|' + r
            res = res + ')'
            return {'res': res, 'pattern_list': pattern_list}
        else:
            return {
                'res': r'(?:' + self.singleton_regex[key_list] + r')',
                'pattern_list': [key_list] if key_list[0] == '%' else []
            }





import unittest, json
from etk.timeseries_processor import TimeseriesProcessor
from etk.etk import ETK
from etk.knowledge_graph_schema import KGSchema

kg_schema = KGSchema(json.load(open('etk/unit_tests/ground_truth/test_config.json')))

etk = ETK(kg_schema=kg_schema)


class TestTimeseriesProcessor(unittest.TestCase):
    def test_excel_file(self) -> None:
        annotation = 'etk/timeseries/DIESEL_june_annotation.json'
        spreadsheet = 'etk/unit_tests/ground_truth/DIESEL_june_2017.xlsx'

        timeseriesProcessor = TimeseriesProcessor(etk=etk, annotation=annotation, spreadsheet=spreadsheet)
        docs = [doc.cdr_document for doc in timeseriesProcessor.timeseries_extractor()]
        selected_docs = docs[1]
        expected_metadata = {
            "name": "AVERAGE DIESEL (AUTOMATIVE GAS OIL) PRICES/ Litre NGN",
            "granularity":"monthly",
            "provenance": {
                "filename": "DIESEL_june_2017.xlsx",
                "sheet": 0,
                "row": 5
            },
            "location": "Abuja",
            "yearly_change": "23.776223776223777",
            "monthly_change": "6.3701923076923075",
            "uid": "5e8c87d3d291944f01f6084c5dee223ee31746e2"
        }
        expected_ts = [('2016-09-01', 189.5),
                       ('2016-10-01', 180),
                       ('2016-11-01', 182.5),
                       ('2016-12-01', 180),
                       ('2017-01-01', 270),
                       ('2017-02-01', 252),
                       ('2017-03-01', 255),
                       ('2017-04-01', 248),
                       ('2017-05-01', 208),
                       ('2017-06-01', 221.25)]
        actual_ts = []
        for t in selected_docs['ts']:
             if 'instant' in t['time'].keys():
                  actual_ts.append((t['time']['instant'], t['value']))
             else:
                  actual_ts.append((t['time']['span']['start_time'], t['value']))
     
        self.assertEqual(actual_ts, expected_ts)
        for k in expected_metadata.keys():
            self.assertEqual(selected_docs['metadata'][k], expected_metadata[k])



import unittest
from etk.extractors.excel_extractor import ExcelExtractor
import datetime

ee = ExcelExtractor()
file_path = 'etk/unit_tests/ground_truth/NST-Main Sheet.xlsx'
sheet_name = 'NST Main Dataset'


class TestExcelExtractor(unittest.TestCase):
    def test_cur_col_cur_row(self) -> None:
        extractions = ee.extract(file_name=file_path,
                                sheet_name=sheet_name,
                                region=['A,2', 'B,10'],
                                variables={
                                    'value': '$col,$row',
                                    })
        expected = [{'value': "Army Barracks Attack Ahead of Goodluck Jonathan's Inauguration"},
                    {'value': 'Kaduna'},
                    {'value': 'Morning Attack in Zaria'},
                    {'value': "Shehu of Borno's Brother Killed"},
                    {'value': 'Attempted Attack on Government Vaccine Warehouse'},
                    {'value': 'Militants Kill Muslim Cleric Ibrahim Burkuti'},
                    {'value': "Bomb Blasts Outside St. Patrick's Catholic Church"},
                    {'value': 'Bombing of Kano Motor Park'}
                    ]

        self.assertEqual(extractions, expected)

    def test_const_col(self) -> None:
        extractions = ee.extract(file_name=file_path,
                               sheet_name=sheet_name,
                               region=['A,2', 'B,10'],
                               variables={
                                   'value': '$B'})
        expected = [{'value': 1},
                    {'value': 1},
                    {'value': 1},
                    {'value': 1},
                    {'value': 1},
                    {'value': 1},
                    {'value': 1},
                    {'value': 1}
                    ]

        self.assertEqual(extractions, expected)

    def test_const_col_const_row(self) -> None:
        extractions = ee.extract(file_name=file_path,
                                sheet_name=sheet_name,
                                region=['A,2', 'B,10'],
                                variables={
                                    'value': '$A,$10',
                                    })
        expected = [{'value': 'David Usman and  Shot Dead'},
                    {'value': 'David Usman and  Shot Dead'},
                    {'value': 'David Usman and  Shot Dead'},
                    {'value': 'David Usman and  Shot Dead'},
                    {'value': 'David Usman and  Shot Dead'},
                    {'value': 'David Usman and  Shot Dead'},
                    {'value': 'David Usman and  Shot Dead'},
                    {'value': 'David Usman and  Shot Dead'}
                    ]

        self.assertEqual(extractions, expected)

    def test_offset_col_const_row(self) -> None:
        extractions = ee.extract(file_name=file_path,
                                sheet_name=sheet_name,
                                region=['A,2', 'B,10'],
                                variables={
                                    'value': '$A+1,$row',
                                    })
        expected = [{'value': datetime.date(2011, 5, 29)},
                    {'value': datetime.date(2011, 5, 29)},
                    {'value': datetime.date(2011, 5, 30)},
                    {'value': datetime.date(2011, 5, 30)},
                    {'value': datetime.date(2011, 6, 2)},
                    {'value': datetime.date(2011, 6, 6)},
                    {'value': datetime.date(2011, 6, 7)},
                    {'value': datetime.date(2011, 6, 7)}
                    ]

        self.assertEqual(extractions, expected)

    def test_cur_col_const_row(self) -> None:
        extractions = ee.extract(file_name=file_path,
                                 sheet_name=sheet_name,
                                 region=['A,2', 'C,10'],
                                 variables={
                                     'value': '$col,$10',
                                 })

        expected = [{'value': 'David Usman and  Shot Dead'},
                    {'value': datetime.date(2011, 6, 7)},
                    {'value': 'David Usman and  Shot Dead'},
                    {'value': datetime.date(2011, 6, 7)},
                    {'value': 'David Usman and  Shot Dead'},
                    {'value': datetime.date(2011, 6, 7)},
                    {'value': 'David Usman and  Shot Dead'},
                    {'value': datetime.date(2011, 6, 7)},
                    {'value': 'David Usman and  Shot Dead'},
                    {'value': datetime.date(2011, 6, 7)},
                    {'value': 'David Usman and  Shot Dead'},
                    {'value': datetime.date(2011, 6, 7)},
                    {'value': 'David Usman and  Shot Dead'},
                    {'value': datetime.date(2011, 6, 7)},
                    {'value': 'David Usman and  Shot Dead'},
                    {'value': datetime.date(2011, 6, 7)}]

        self.assertEqual(extractions, expected)

    def test_const_col_cur_row(self) -> None:
        extractions = ee.extract(file_name=file_path,
                                 sheet_name=sheet_name,
                                 region=['A,2', 'B,10'],
                                 variables={
                                     'value': '$A,$row',
                                 })
        expected = [{'value': "Army Barracks Attack Ahead of Goodluck Jonathan's Inauguration"},
                    {'value': 'Kaduna'},
                    {'value': 'Morning Attack in Zaria'},
                    {'value': "Shehu of Borno's Brother Killed"},
                    {'value': 'Attempted Attack on Government Vaccine Warehouse'},
                    {'value': 'Militants Kill Muslim Cleric Ibrahim Burkuti'},
                    {'value': "Bomb Blasts Outside St. Patrick's Catholic Church"},
                    {'value': 'Bombing of Kano Motor Park'}]

        self.assertEqual(extractions, expected)

    # def test_col_name_to_num(self) -> None:
    #     self.assertEquals(ee.col_name_to_num('A'), 0)
    #     self.assertEquals(ee.col_name_to_num('Z'), 25)
    #     self.assertEquals(ee.col_name_to_num('AA'), 26)
    #     self.assertEquals(ee.col_name_to_num('aA'), 26)
    #     self.assertEquals(ee.col_name_to_num('aa'), 26)
    #
    # def test_row_name_to_name(self) -> None:
    #     self.assertEquals(ee.row_name_to_num('1'), 0)
    #     self.assertEquals(ee.row_name_to_num('100'), 99)
    #     self.assertRaises(ValueError, ee.row_name_to_num, 'abc123')
    #     self.assertRaises(ValueError, ee.row_name_to_num, '0')
    #
    # def test_parse_variable(self) -> None:
    #     self.assertEquals(ee.parse_variable('$A', 10, 10), (0,))
    #     self.assertEquals(ee.parse_variable('$A+2,$1', 10, 10), (0, 2))
    #     self.assertEquals(ee.parse_variable('2+$A,$1', 10, 10), (0, 2))
    #     self.assertEquals(ee.parse_variable('$col,$1', 10, 10), (0, 10))
    #     self.assertEquals(ee.parse_variable('$A,$row', 10, 10), (10, 0))
    #     self.assertEquals(ee.parse_variable('$A+2,$row', 10, 10), (10, 2))


if __name__ == '__main__':
    unittest.main()



from etk.extractors.regex_extractor import RegexExtractor


class HostnameExtractor(RegexExtractor):
    """
    **Description**
        This class inherits RegexExtractor by predefining the regex pattern for hostname

    Examples:
        ::

            hostname_extractor = HostnameExtractor()
            hostname_extractor.extract(text=input_doc)

    """
    def __init__(self):
        hostname_pattern = r"\b(?:[a-zA-Z0-9](?:[a-zA-Z0-9\-]{,61}[a-zA-Z0-9])?\.)+" \
                           r"(?!html|php|jsp|xml|pdf|asp|css|aspx|phtml)[a-zA-Z]{2,6}\b"
        RegexExtractor.__init__(self, pattern=hostname_pattern, extractor_name="hostname extractor")



from etk.extraction import Extractable, Extraction
from typing import List, Dict
from etk.etk_exceptions import StoreExtractionError
from etk.storage_provenance_record import StorageProvenanceRecord
from numbers import Number


class Segment(Extractable):
    """
    An individual segment in a document.
    For now, it supports recording of JSONPath results, but we should consider extending
    to record segments within a text doc, e.g., by start and end char, or segments within
    a token list with start and end tokens.
    """

    def __init__(self, json_path: str, _value: Dict, _document) -> None:
        Extractable.__init__(self)
        self.json_path = json_path
        self._value = _value
        self._extractions = dict()
        self._document = _document

    @property
    def full_path(self) -> str:
        """
        Returns: The full path of a JSONPath match
        """
        return self.json_path

    @property
    def document(self):
        """
        Still thinking about this, having the parent doc inside each extractable is convenient to avoid
        passing around the parent doc to all methods.

        Returns: the parent Document
        """
        return self._document

    def store(self, extractions: List[Extraction], attribute: str, group_by_tags: bool = True) -> None:
        """
        Records extractions in the container, and for each individual extraction inserts a
        ProvenanceRecord to record where the extraction is stored.
        Records the "output_segment" in the provenance.

        Extractions are always recorded in a list.

        Errors out if the segment is primitive, such as a string.

        Args:
            extractions (List[Extraction]):
            attribute (str): where to store the extractions.
            group_by_tags (bool): Set to True to use tags as sub-keys, and values of Extractions
                with the same tag will be stored in a list as the value of the corresponding key.
                (if none of the Extractions has 'tag', do not group by tags)

        Returns:

        """
        if not isinstance(self._value, dict):
            raise StoreExtractionError("segment is type: " + str(type(self._value)))

        if not len(extractions):
            return

        if group_by_tags:
            try:
                next(x for x in extractions if x.tag)  # if there is at least one extraction with a tag
                if attribute not in self._extractions:
                    self._extractions[attribute] = set([])
                    self._value[attribute] = {}
                extraction_provenances = {}
                for e in extractions:
                    tag = e.tag if e.tag else 'NO_TAGS'
                    if tag not in self.value[attribute]:
                        self.value[attribute][tag] = [e.value]
                    else:
                        if e.value not in self.value[attribute][tag]:
                            self.value[attribute][tag].append(e.value)
                    # TODO: handle provenance of non literals
                    if isinstance(e.value, Number) or isinstance(e.value, str):
                        extraction_provenances[e.value] = e.prov_id
                self._extractions[attribute] = self._extractions[attribute].union(extractions)
                new_id = self._document.provenance_id_index  # for the purpose of provenance hierarrchy tracking
                storage_provenance_record: StorageProvenanceRecord = StorageProvenanceRecord(new_id, self.json_path,
                                                                                             attribute,
                                                                                             extraction_provenances,
                                                                                             self.document)
                self._document.provenance_id_index_incrementer()
                self._document.provenances[new_id] = storage_provenance_record
                self.create_storage_provenance(storage_provenance_record)
                return
            except StopIteration:
                pass

        if attribute not in self._extractions:
            self._extractions[attribute] = set([])
            self._value[attribute] = list()

        self._extractions[attribute] = self._extractions[attribute].union(extractions)
        extraction_provenances = dict()
        for a_extraction in extractions:
            # TODO: handle provenance of non literals
            if isinstance(a_extraction.value, Number) or isinstance(a_extraction.value, str):
                extraction_provenances[a_extraction.value] = a_extraction.prov_id
            if a_extraction.value not in self._value[attribute]:
                self._value[attribute].append(a_extraction.value)

        new_id = self._document.provenance_id_index  # for the purpose of provenance hierarchy tracking
        storage_provenance_record: StorageProvenanceRecord = StorageProvenanceRecord(new_id, self.json_path, attribute,
                                                                                     extraction_provenances,
                                                                                     self.document)
        self._document.provenance_id_index_incrementer()
        self._document.provenances[new_id] = storage_provenance_record
        self.create_storage_provenance(storage_provenance_record)

    @property
    def extractions(self) -> Dict:
        """
        Get the extractions stored in this container.
        Returns: Dict

        """
        return self._extractions

    def create_storage_provenance(self, storage_provenance_record: StorageProvenanceRecord) -> None:
        if "provenances" not in self.document.cdr_document:
            self.document.cdr_document["provenances"] = list()
        self.document.cdr_document["provenances"].append(self.get_dict_storage_provenance(storage_provenance_record))

    def get_dict_storage_provenance(self, storage_provenance_record: StorageProvenanceRecord):
        prov_dict = dict()
        prov_dict["@id"] = storage_provenance_record.id
        prov_dict["@type"] = "storage_provenance_record"
        prov_dict["doc_id"] = storage_provenance_record.doc_id
        prov_dict["field"] = storage_provenance_record.field
        jsonpath = storage_provenance_record.destination
        prov_dict["destination"] = jsonpath
        if jsonpath in self._document.jsonpath_provenances:
            self._document.jsonpath_provenances[jsonpath].append(storage_provenance_record.id)
        else:
            self._document.jsonpath_provenances[jsonpath] = [storage_provenance_record.id]
            prov_dict["parent_provenances"] = storage_provenance_record.extraction_provenances
        return prov_dict



import re
import chardet

def get_encoding(page):
    # Regex for XML and HTML Meta charset declaration
    charset_re = re.compile(r'<meta.*?charset=["\']*(.+?)["\'>]', flags=re.I)
    pragma_re = re.compile(r'<meta.*?content=["\']*;?charset=(.+?)["\'>]', flags=re.I)
    xml_re = re.compile(r'^<\?xml.*?encoding=["\']*(.+?)["\'>]')

    declared_encodings = (charset_re.findall(page) +
            pragma_re.findall(page) +
            xml_re.findall(page))

    # Try any declared encodings
    if len(declared_encodings) > 0:
        for declared_encoding in declared_encodings:
            try:
                page.decode(custom_decode(declared_encoding))
                return custom_decode(declared_encoding)
            except UnicodeDecodeError:
                pass

    # Fallback to chardet if declared encodings fail
    text = re.sub('</?[^>]*>\s*', ' ', page)
    enc = 'utf-8'
    if not text.strip() or len(text) < 10:
        return enc # can't guess
    res = chardet.detect(text)
    enc = res['encoding'] or 'utf-8'
    #print '->', enc, "%.2f" % res['confidence']
    enc = custom_decode(enc)
    return enc

def custom_decode(encoding):
    """Overrides encoding when charset declaration
       or charset determination is a subset of a larger
       charset.  Created because of issues with Chinese websites"""
    encoding = encoding.lower()
    alternates = {
        'big5': 'big5hkscs',
        'gb2312': 'gb18030',
        'ascii': 'utf-8',
        'MacCyrillic': 'cp1251',
    }
    if encoding in alternates:
        return alternates[encoding]
    else:
        return encoding



import unittest, json
from etk.csv_processor import CsvProcessor
from etk.etk import ETK
from etk.knowledge_graph_schema import KGSchema

csv_str = """text,with,Polish,non-Latin,lettes
1,2,3,4,5,6
a,b,c,d,e,f

gęś,zółty,wąż,idzie,wąską,dróżką,
,b,c,s,w,f
"""

kg_schema = KGSchema(json.load(open('etk/unit_tests/ground_truth/test_config.json')))

etk = ETK(kg_schema=kg_schema)


class TestCsvProcessor(unittest.TestCase):
    def test_csv_str_with_all_args(self) -> None:
        csv_processor = CsvProcessor(etk=etk,
                                     heading_row=1,
                                     content_start_row=2,
                                     heading_columns=(1, 3),
                                     content_end_row=3,
                                     ends_with_blank_row=True,
                                     remove_leading_empty_rows=True,
                                     required_columns=['text'])

        test_docs = [doc.cdr_document for doc in
                     csv_processor.tabular_extractor(table_str=csv_str, dataset='test_csv_str_with_all_args')]

        expected_docs = [{'text': '1', 'with': '2', 'Polish': '3', 'non-Latin': '4', 'lettes': '5',
                          'dataset': 'test_csv_str_with_all_args'},
                         {'text': 'a', 'with': 'b', 'Polish': 'c', 'non-Latin': 'd', 'lettes': 'e',
                          'dataset': 'test_csv_str_with_all_args'}]

        self.assertEqual(test_docs, expected_docs)

    def test_csv_str_with_ends_with_blank_row_false(self) -> None:
        csv_processor = CsvProcessor(etk=etk,
                                     heading_row=1,
                                     content_start_row=2,
                                     heading_columns=(1, 3),
                                     content_end_row=4,
                                     ends_with_blank_row=False,
                                     remove_leading_empty_rows=True,
                                     required_columns=['text'])

        test_docs = [doc.cdr_document for doc in
                     csv_processor.tabular_extractor(table_str=csv_str, dataset='test_set')]

        expected_docs = [
            {'text': '1', 'with': '2', 'Polish': '3', 'non-Latin': '4', 'lettes': '5', 'dataset': 'test_set'},
            {'text': 'a', 'with': 'b', 'Polish': 'c', 'non-Latin': 'd', 'lettes': 'e', 'dataset': 'test_set'}]

        self.assertEqual(test_docs, expected_docs)

    def test_csv_file_with_no_header(self) -> None:
        csv_processor = CsvProcessor(etk=etk,
                                     content_start_row=1,
                                     content_end_row=4,
                                     ends_with_blank_row=False,
                                     remove_leading_empty_rows=True)
        filename = 'etk/unit_tests/ground_truth/sample_csv.csv'

        test_docs = [doc.cdr_document for doc in
                     csv_processor.tabular_extractor(filename=filename, dataset='test_set')]

        expected_docs = [{'C0': 'name1', 'C1': 'name2', 'file_name': 'etk/unit_tests/ground_truth/sample_csv.csv',
                          'dataset': 'test_set'}]

        self.assertEqual(test_docs, expected_docs)

    def test_csv_file_with_no_header_not_ends_with_blank_row(self) -> None:
        csv_processor = CsvProcessor(etk=etk,
                                     content_start_row=1,
                                     content_end_row=8,
                                     ends_with_blank_row=False,
                                     remove_leading_empty_rows=True)
        filename = 'etk/unit_tests/ground_truth/sample_csv.csv'

        test_docs = [doc.cdr_document for doc in
                     csv_processor.tabular_extractor(filename=filename, dataset='test_set')]

        expected_docs = [{'C0': '', 'C1': 'name1', 'C2': 'name2', 'C3': '', 'C4': '',
                          'file_name': 'etk/unit_tests/ground_truth/sample_csv.csv', 'dataset': 'test_set'},
                         {'C0': 'col11', 'C1': 'col12', 'C2': 'col13', 'C3': '', 'C4': 'col15',
                          'file_name': 'etk/unit_tests/ground_truth/sample_csv.csv', 'dataset': 'test_set'},
                         {'C0': 'col21', 'C1': 'col22', 'C2': 'col23', 'C3': 'col24', 'C4': 'col25',
                          'file_name': 'etk/unit_tests/ground_truth/sample_csv.csv', 'dataset': 'test_set'},
                         {'C0': 'col31', 'C1': 'col32', 'C2': 'col33', 'C3': 'col34', 'C4': 'col35',
                          'file_name': 'etk/unit_tests/ground_truth/sample_csv.csv', 'dataset': 'test_set'}]

        self.assertEqual(test_docs, expected_docs)

    def test_real_csv_file_1(self) -> None:
        csv_processor = CsvProcessor(etk=etk,
                                     heading_row=1,
                                     content_end_row=4,
                                     ends_with_blank_row=False)

        file_path = 'etk/unit_tests/ground_truth/acled_raw_data.csv'

        test_docs = [doc.cdr_document for doc in
                     csv_processor.tabular_extractor(filename=file_path, dataset='test_set')]

        expected_docs = [{'data_id': 336907, 'iso': 180, 'event_id_cnty': 'DRC11776', 'event_id_no_cnty': 11776,
                          'event_date': '2018-01-13', 'year': 2018, 'time_precision': 1, 'event_type':
                              'Battle-No change of territory',
                          'actor1': 'Military Forces of Democratic Republic of Congo (2001-)',
                          'assoc_actor_1': '', 'inter1': 1, 'actor2': 'ADF: Allied Democratic Forces',
                          'assoc_actor_2': '',
                          'inter2': 2, 'interaction': 12, 'region': 'Central Africa',
                          'country': 'Democratic Republic of Congo',
                          'admin1': 'Nord-Kivu', 'admin2': 'Nord-Kivu', 'admin3': 'Oicha', 'location': 'Oicha',
                          'latitude': '0.7',
                          'longitude': '29.5167', 'geo_precision': 1, 'source': 'Radio Okapi',
                          'source_scale': 'Subnational',
                          'notes': "Presumed FARDC attacked the ADF in the periphery of Oicha on January 13th. Shots were heard "
                                   "in the locality and it is suspected that the FARDC are attacking the 'death triangle' situated in between "
                                   "Mbau, Kamango and Eringeti. The reports are not confirmed by the military.",
                          'fatalities': 0,
                          'timestamp': 1516117305, 'file_name': 'etk/unit_tests/ground_truth/acled_raw_data.csv',
                          'dataset': 'test_set'},
                         {'data_id': 336908, 'iso': 180, 'event_id_cnty': 'DRC11777', 'event_id_no_cnty': 11777,
                          'event_date': '2018-01-13', 'year': 2018, 'time_precision': 1,
                          'event_type': 'Battle-No change of territory',
                          'actor1': 'Military Forces of Democratic Republic of Congo (2001-)', 'assoc_actor_1': '',
                          'inter1': 1,
                          'actor2': 'ADF: Allied Democratic Forces', 'assoc_actor_2': '', 'inter2': 2,
                          'interaction': 12,
                          'region': 'Central Africa', 'country': 'Democratic Republic of Congo', 'admin1': 'Nord-Kivu',
                          'admin2': 'Beni', 'admin3': 'Beni', 'location': 'Beni', 'latitude': '0.49658',
                          'longitude': '29.4654',
                          'geo_precision': 1, 'source': 'Reuters; Radio Okapi', 'source_scale': 'Subnational',
                          'notes': 'The FARDC launched, on January 13th, an offensive against the ADF in Beni and Lubero, '
                                   'in response to the recents attacks by the group. Gunfires and explosions were heard '
                                   'in Beni all throughout Saturday (13th).', 'fatalities': 0, 'timestamp': 1516117305,
                          'file_name': 'etk/unit_tests/ground_truth/acled_raw_data.csv', 'dataset': 'test_set'},
                         {'data_id': 336909, 'iso': 180, 'event_id_cnty': 'DRC11778', 'event_id_no_cnty': 11778,
                          'event_date': '2018-01-13', 'year': 2018, 'time_precision': 1,
                          'event_type': 'Battle-No change '
                                        'of territory',
                          'actor1': 'Military Forces of Democratic Republic of Congo (2001-)',
                          'assoc_actor_1': '', 'inter1': 1, 'actor2': 'ADF: Allied Democratic Forces',
                          'assoc_actor_2': '',
                          'inter2': 2, 'interaction': 12, 'region': 'Central Africa',
                          'country': 'Democratic Republic of Congo',
                          'admin1': 'Nord-Kivu', 'admin2': 'Nord-Kivu', 'admin3': 'Lubero', 'location': 'Lubero',
                          'latitude': '-0.15867', 'longitude': '29.2386', 'geo_precision': 1,
                          'source': 'Reuters; Radio Okapi',
                          'source_scale': 'Subnational', 'notes': 'The FARDC launched, on January 13th, an offensive '
                                                                  'against the ADF in Beni and Lubero, in response to the recents attacks by the group.',
                          'fatalities': 0, 'timestamp': 1516117305,
                          'file_name': 'etk/unit_tests/ground_truth/acled_raw_data.csv',
                          'dataset': 'test_set'}]

        self.assertEqual(test_docs, expected_docs)

    def test_real_csv_file_2(self) -> None:
        csv_processor = CsvProcessor(etk=etk,
                                     heading_row=2,
                                     content_start_row=10,
                                     content_end_row=13)

        file_path = 'etk/unit_tests/ground_truth/masie_4km_allyears_extent_sqkm.csv'

        test_docs = [doc.cdr_document for doc in
                     csv_processor.tabular_extractor(filename=file_path, dataset='test_set')]

        expected_docs = [
            {'yyyyddd': 2006008, ' (0) Northern_Hemisphere': '13536736.84', ' (1) Beaufort_Sea': ' 1069710.81',
             ' (2) Chukchi_Sea': '  966006.16', ' (3) East_Siberian_Sea': ' 1087102.72',
             ' (4) Laptev_Sea': '  897773.37',
             ' (5) Kara_Sea': '  927602.17', ' (6) Barents_Sea': '  474574.82', ' (7) Greenland_Sea': '  590029.18',
             ' (8) Baffin_Bay_Gulf_of_St._Lawrence': ' 1005790.38', ' (9) Canadian_Archipelago': '  852715.31',
             ' (10) Hudson_Bay': ' 1260779.00', ' (11) Central_Arctic': ' 3240326.47',
             ' (12) Bering_Sea': '  692832.54',
             ' (13) Baltic_Sea': '   21327.46', ' (14) Sea_of_Okhotsk': '  424563.54',
             ' (15) Yellow_Sea': '   14830.45',
             ' (16) Cook_Inlet': '    8202.95',
             'file_name': 'etk/unit_tests/ground_truth/masie_4km_allyears_extent_sqkm.csv',
             'dataset': 'test_set'},
            {'yyyyddd': 2006009, ' (0) Northern_Hemisphere': '13536887.64', ' (1) Beaufort_Sea': ' 1069710.81',
             ' (2) Chukchi_Sea': '  966006.16', ' (3) East_Siberian_Sea': ' 1087102.72',
             ' (4) Laptev_Sea': '  897773.37',
             ' (5) Kara_Sea': '  927602.17', ' (6) Barents_Sea': '  474574.82', ' (7) Greenland_Sea': '  590029.18',
             ' (8) Baffin_Bay_Gulf_of_St._Lawrence': ' 1005790.38', ' (9) Canadian_Archipelago': '  852715.31',
             ' (10) Hudson_Bay': ' 1260779.00', ' (11) Central_Arctic': ' 3240326.47',
             ' (12) Bering_Sea': '  692832.54',
             ' (13) Baltic_Sea': '   21478.25', ' (14) Sea_of_Okhotsk': '  424563.54',
             ' (15) Yellow_Sea': '   14830.45',
             ' (16) Cook_Inlet': '    8202.95',
             'file_name': 'etk/unit_tests/ground_truth/masie_4km_allyears_extent_sqkm.csv',
             'dataset': 'test_set'},
            {'yyyyddd': 2006010, ' (0) Northern_Hemisphere': '13505426.35', ' (1) Beaufort_Sea': ' 1069710.81',
             ' (2) Chukchi_Sea': '  966006.16', ' (3) East_Siberian_Sea': ' 1087102.72',
             ' (4) Laptev_Sea': '  897773.37',
             ' (5) Kara_Sea': '  933999.29', ' (6) Barents_Sea': '  448185.27', ' (7) Greenland_Sea': '  588279.64',
             ' (8) Baffin_Bay_Gulf_of_St._Lawrence': ' 1016857.87', ' (9) Canadian_Archipelago': '  852715.31',
             ' (10) Hudson_Bay': ' 1260779.00', ' (11) Central_Arctic': ' 3217380.82',
             ' (12) Bering_Sea': '  705348.17',
             ' (13) Baltic_Sea': '   21493.81', ' (14) Sea_of_Okhotsk': '  414191.19',
             ' (15) Yellow_Sea': '   14830.45',
             ' (16) Cook_Inlet': '    8202.95',
             'file_name': 'etk/unit_tests/ground_truth/masie_4km_allyears_extent_sqkm.csv',
             'dataset': 'test_set'},
            {'yyyyddd': 2006011, ' (0) Northern_Hemisphere': '13493030.93', ' (1) Beaufort_Sea': ' 1069710.81',
             ' (2) Chukchi_Sea': '  966006.16', ' (3) East_Siberian_Sea': ' 1087102.72',
             ' (4) Laptev_Sea': '  897773.37',
             ' (5) Kara_Sea': '  933999.29', ' (6) Barents_Sea': '  448185.27', ' (7) Greenland_Sea': '  588279.64',
             ' (8) Baffin_Bay_Gulf_of_St._Lawrence': ' 1013091.28', ' (9) Canadian_Archipelago': '  852715.31',
             ' (10) Hudson_Bay': ' 1260779.00', ' (11) Central_Arctic': ' 3217380.82',
             ' (12) Bering_Sea': '  696719.33',
             ' (13) Baltic_Sea': '   21493.81', ' (14) Sea_of_Okhotsk': '  414191.19',
             ' (15) Yellow_Sea': '   14830.45',
             ' (16) Cook_Inlet': '    8202.95',
             'file_name': 'etk/unit_tests/ground_truth/masie_4km_allyears_extent_sqkm.csv',
             'dataset': 'test_set'}]

        self.assertEqual(test_docs, expected_docs)

    def test_tab_file(self):
        csv_processor = CsvProcessor(etk=etk,
                                     heading_row=1)
        test_str = 'Event ID\tEvent Date\tSource Name\tSource Sectors\tSource Country\tEvent Text\tCAMEO Code\tIntensity\tTarget Name\tTarget Sectors\tTarget Country\tStory ID\tSentence Number\tPublisher\tCity\tDistrict\tProvince\tCountry\tLatitude\tLongitude\n926685\t1995-01-01\tExtremist (Russia)\tRadicals / Extremists / Fundamentalists,Dissident\tRussian Federation\tPraise or endorse\t051\t3.4\tBoris Yeltsin\tElite,Executive,Executive Office,Government\tRussian Federation\t28235806\t5\tThe Toronto Star\tMoscow\t\tMoskva\tRussian Federation\t55.7522\t37.6156\n926687\t1995-01-01\tGovernment (Bosnia and Herzegovina)\tGovernment\tBosnia and Herzegovina\tExpress intent to cooperate\t030\t4\tCitizen (Serbia)\tGeneral Population / Civilian / Social,Social\tSerbia\t28235807\t1\tThe Toronto Star\t\t\tBosnia\tBosnia and Herzegovina\t44\t18\n926686\t1995-01-01\tCitizen (Serbia)\tGeneral Population / Civilian / Social,Social\tSerbia\tExpress intent to cooperate\t030\t4\tGovernment (Bosnia and Herzegovina)\tGovernment\tBosnia and Herzegovina\t28235807\t1\tThe Toronto Star\t\t\tBosnia\tBosnia and Herzegovina\t44\t18\n926688\t1995-01-01\tCanada\t\tCanada\tPraise or endorse\t051\t3.4\tCity Mayor (Canada)\tGovernment,Local,Municipal\tCanada\t28235809\t3\tThe Toronto Star\t\t\tOntario\tCanada\t49.2501\t-84.4998\n'
        test_docs = [doc.cdr_document for doc in
                     csv_processor.tabular_extractor(filename='testfile.tab', file_type='tsv', file_content=test_str,
                                                     dataset='test_set')]
        expected_docs = [{'Event ID': 926685, 'Event Date': '1995-01-01', 'Source Name': 'Extremist (Russia)',
                          'Source Sectors': 'Radicals / Extremists / Fundamentalists,Dissident',
                          'Source Country': 'Russian Federation', 'Event Text': 'Praise or endorse',
                          'CAMEO Code': '051', 'Intensity': '3.4', 'Target Name': 'Boris Yeltsin',
                          'Target Sectors': 'Elite,Executive,Executive Office,Government',
                          'Target Country': 'Russian Federation', 'Story ID': 28235806, 'Sentence Number': 5,
                          'Publisher': 'The Toronto Star', 'City': 'Moscow', 'District': '', 'Province': 'Moskva',
                          'Country': 'Russian Federation', 'Latitude': '55.7522', 'Longitude': '37.6156',
                          'file_name': 'testfile.tab', 'dataset': 'test_set'},
                         {'Event ID': 926687, 'Event Date': '1995-01-01',
                          'Source Name': 'Government (Bosnia and Herzegovina)', 'Source Sectors': 'Government',
                          'Source Country': 'Bosnia and Herzegovina', 'Event Text': 'Express intent to cooperate',
                          'CAMEO Code': '030', 'Intensity': 4, 'Target Name': 'Citizen (Serbia)',
                          'Target Sectors': 'General Population / Civilian / Social,Social', 'Target Country': 'Serbia',
                          'Story ID': 28235807, 'Sentence Number': 1, 'Publisher': 'The Toronto Star', 'City': '',
                          'District': '', 'Province': 'Bosnia', 'Country': 'Bosnia and Herzegovina', 'Latitude': 44,
                          'Longitude': 18, 'file_name': 'testfile.tab', 'dataset': 'test_set'},
                         {'Event ID': 926686, 'Event Date': '1995-01-01', 'Source Name': 'Citizen (Serbia)',
                          'Source Sectors': 'General Population / Civilian / Social,Social', 'Source Country': 'Serbia',
                          'Event Text': 'Express intent to cooperate', 'CAMEO Code': '030', 'Intensity': 4,
                          'Target Name': 'Government (Bosnia and Herzegovina)', 'Target Sectors': 'Government',
                          'Target Country': 'Bosnia and Herzegovina', 'Story ID': 28235807, 'Sentence Number': 1,
                          'Publisher': 'The Toronto Star', 'City': '', 'District': '', 'Province': 'Bosnia',
                          'Country': 'Bosnia and Herzegovina', 'Latitude': 44, 'Longitude': 18,
                          'file_name': 'testfile.tab', 'dataset': 'test_set'},
                         {'Event ID': 926688, 'Event Date': '1995-01-01', 'Source Name': 'Canada', 'Source Sectors': '',
                          'Source Country': 'Canada', 'Event Text': 'Praise or endorse', 'CAMEO Code': '051',
                          'Intensity': '3.4', 'Target Name': 'City Mayor (Canada)',
                          'Target Sectors': 'Government,Local,Municipal', 'Target Country': 'Canada',
                          'Story ID': 28235809, 'Sentence Number': 3, 'Publisher': 'The Toronto Star', 'City': '',
                          'District': '', 'Province': 'Ontario', 'Country': 'Canada', 'Latitude': '49.2501',
                          'Longitude': '-84.4998', 'file_name': 'testfile.tab', 'dataset': 'test_set'}]
        self.assertEqual(test_docs, expected_docs)

    def test_real_excel_with_sheetname(self) -> None:
        csv_processor = CsvProcessor(etk=etk,
                                     heading_row=1,
                                     content_start_row=10,
                                     content_end_row=12)

        file_path = 'etk/unit_tests/ground_truth/NST-Main Sheet.xlsx'

        sheets_name = 'NST Main Dataset'

        test_docs = [doc.cdr_document for doc in
                     csv_processor.tabular_extractor(filename=file_path, sheet_name=sheets_name, dataset='test_set')]

        expected_docs = [{'Title': 'David Usman and  Shot Dead', 'Date': '2011-06-07',
                          'Community (city,town, ward)': 'Maiduguri', 'LGA': 'Maiduguri', 'State': 'Borno',
                          'Total Deaths': 2,
                          'Boko Haram (P)': 'Boko Haram', 'State Actor (P)': '',
                          'Sectarian Actor (excluding BH) (P)': '',
                          'Other Armed Actor (P)': '', 'Kidnapper (P)': '', 'Robber (P)': '', 'Other (P)': '',
                          'Election-related Actor (P)': '', 'Cameroon State Actor (P)': '', 'Boko Haram (V)': '',
                          'State Actor (V)': '',
                          'Sectarian Actor (V)': 2, 'Other Armed Actor (V)': '', 'Political Actor (V)': '',
                          'Kidnapper (V)': '',
                          'Kidnapee (V)': '', 'Robber (V)': '', 'Journalist (V)': '', 'Civilian (V)': '',
                          'Election-related Actor (V)': '',
                          'Cameroon State Actor': '', 'Bomb': '', 'Gun': 'Gun', 'Machete': '', 'Suicide Bombing': '',
                          'Other Weapon': '', 'TK': 'Targeted Killing', 'Drinking Establishment': '',
                          'Goverment Building': '',
                          'Church': '', 'Mosque': '', 'Bank': '', 'School': '', 'Other Location': 'Other', 'Notes': '',
                          'Sources 1': 'http://allafrica.com/stories/201106100373.html',
                          'Sources 2': 'http://www.bbc.co.uk/news/world-africa-13724349',
                          'Sources 3': '', 'Latitude': '', 'Longitude': '',
                          'full place name': 'Maiduguri, Borno, Nigeria',
                          'country': 'Nigeria', 'file_name': 'etk/unit_tests/ground_truth/NST-Main Sheet.xlsx',
                          'dataset': 'test_set'},
                         {'Title': 'Explosion, Firefight at Gwange Police Station', 'Date': '2011-06-07',
                          'Community (city,town, ward)': 'Gwange Police Station, Maiduguri', 'LGA': 'Maiduguri',
                          'State': 'Borno',
                          'Total Deaths': 3, 'Boko Haram (P)': 'Boko Haram', 'State Actor (P)': 'State Actor',
                          'Sectarian Actor (excluding BH) (P)': '', 'Other Armed Actor (P)': '', 'Kidnapper (P)': '',
                          'Robber (P)': '', 'Other (P)': '', 'Election-related Actor (P)': '',
                          'Cameroon State Actor (P)': '',
                          'Boko Haram (V)': 3, 'State Actor (V)': '', 'Sectarian Actor (V)': '',
                          'Other Armed Actor (V)': '',
                          'Political Actor (V)': '', 'Kidnapper (V)': '', 'Kidnapee (V)': '', 'Robber (V)': '',
                          'Journalist (V)': '',
                          'Civilian (V)': '', 'Election-related Actor (V)': '', 'Cameroon State Actor': '',
                          'Bomb': 'Bomb',
                          'Gun': 'Gun', 'Machete': '', 'Suicide Bombing': '', 'Other Weapon': '', 'TK': '',
                          'Drinking Establishment': '',
                          'Goverment Building': 'Government Building', 'Church': '', 'Mosque': '', 'Bank': '',
                          'School': '',
                          'Other Location': '', 'Notes': '',
                          'Sources 1': 'http://www.google.com/hostednews/afp/article/ALeqM5hofvKayKKAFFtiX9-Ic5bG2ptVmg?docId=CNG.fafcacea0287fbeab90256732f165e1e.771',
                          'Sources 2': 'http://news.xinhuanet.com/english2010/world/2011-06/08/c_13915959.htm',
                          'Sources 3': '',
                          'Latitude': '', 'Longitude': '', 'full place name': 'Maiduguri, Borno, Nigeria',
                          'country': 'Nigeria',
                          'file_name': 'etk/unit_tests/ground_truth/NST-Main Sheet.xlsx', 'dataset': 'test_set'},
                         {'Title': 'Explosions at Dandal Police Station', 'Date': '2011-06-07',
                          'Community (city,town, ward)': 'Dandal Police Station, Maiduguri', 'LGA': 'Maiduguri',
                          'State': 'Borno', 'Total Deaths': 0, 'Boko Haram (P)': 'Boko Haram', 'State Actor (P)': '',
                          'Sectarian Actor (excluding BH) (P)': '', 'Other Armed Actor (P)': '', 'Kidnapper (P)': '',
                          'Robber (P)': '', 'Other (P)': '', 'Election-related Actor (P)': '',
                          'Cameroon State Actor (P)': '',
                          'Boko Haram (V)': '', 'State Actor (V)': 0, 'Sectarian Actor (V)': '',
                          'Other Armed Actor (V)': '',
                          'Political Actor (V)': '', 'Kidnapper (V)': '', 'Kidnapee (V)': '', 'Robber (V)': '',
                          'Journalist (V)': '', 'Civilian (V)': '', 'Election-related Actor (V)': '',
                          'Cameroon State Actor': '',
                          'Bomb': 'Bomb', 'Gun': '', 'Machete': '', 'Suicide Bombing': '', 'Other Weapon': '', 'TK': '',
                          'Drinking Establishment': '', 'Goverment Building': 'Government Building', 'Church': '',
                          'Mosque': '',
                          'Bank': '', 'School': '', 'Other Location': '', 'Notes': '',
                          'Sources 1': 'http://news.xinhuanet.com/english2010/world/2011-06/08/c_13915959.htm',
                          'Sources 2': 'http://www.google.com/hostednews/afp/article/ALeqM5hofvKayKKAFFtiX9-Ic5bG2ptVmg?docId=CNG.fafcacea0287fbeab90256732f165e1e.771',
                          'Sources 3': '', 'Latitude': '', 'Longitude': '',
                          'full place name': 'Maiduguri, Borno, Nigeria',
                          'country': 'Nigeria', 'file_name': 'etk/unit_tests/ground_truth/NST-Main Sheet.xlsx',
                          'dataset': 'test_set'}]

        self.assertEqual(test_docs, expected_docs)

    def test_real_excel_without_sheetname(self) -> None:
        csv_processor = CsvProcessor(etk=etk,
                                     heading_row=1,
                                     content_start_row=10,
                                     content_end_row=12)

        file_path = 'etk/unit_tests/ground_truth/NST-Main Sheet.xlsx'

        test_docs = [doc.cdr_document for doc in
                     csv_processor.tabular_extractor(filename=file_path, dataset='test_set')]

        expected_docs = []

        self.assertEqual(test_docs, expected_docs)

    def test_csv_encoding(self) -> None:
        csv_processor = CsvProcessor(etk=etk,
                                     heading_row=1)
        test_docs = [doc.cdr_document for doc in
                     csv_processor.tabular_extractor(filename='etk/unit_tests/ground_truth/test_encoding.csv',
                                                     dataset='test_set', encoding='utf-16')]
        expected_docs = [{'Country': 'Algeria', 'Category': 'Crude Oil Production', 'DateTime': '2/28/2018 12:00:00 AM',
                          'Close': '1036.0000', 'Frequency': 'Monthly', 'HistoricalDataSymbol': 'ALGERIACRUOILPRO',
                          'LastUpdate': '3/14/2018 2:17:00 PM',
                          'file_name': 'etk/unit_tests/ground_truth/test_encoding.csv', 'dataset': 'test_set'}]

        self.assertEqual(test_docs, expected_docs)


if __name__ == '__main__':
    unittest.main()



from etk.extractors.regex_extractor import RegexExtractor


class BitcoinAddressExtractor(RegexExtractor):
    """
    **Description**
        This class inherits RegexExtractor by predefining the regex pattern based on conditions

    Examples:
        ::

            bitcoin_addr_extractor = BitcoinAddressExtractor(support_Bech32=True)
            bitcoin_addr_extractor.extract(text=input_doc)

    """
    def __init__(self, support_Bech32: bool=False):
        if support_Bech32:
            # a regex support Bech32 type (which is not supported for most applications)
            bitcoin_address_pattern = r"\b[13][a-km-zA-HJ-NP-Z1-9]{25,34}|bc1[a-zA-HJ-NP-Z0-9]{39}|bc1[a-zA-HJ-NP-Z0-9]{59}\b"
        else:
            # simple version supporting P2PKH and P2SH
            bitcoin_address_pattern = r"\b[13][a-km-zA-HJ-NP-Z1-9]{25,34}\b"
        RegexExtractor.__init__(self, pattern=bitcoin_address_pattern, extractor_name="bitcoin address extractor")



import unittest
from etk.extractors.regex_extractor import RegexExtractor, MatchMode


class TestRegexExtractor(unittest.TestCase):
    def test_match_mode_with_group(self) -> None:
        regexp = RegexExtractor('(.)@(.)', 'test_extractor')
        test_str = 'a@1, b@2, c@3, d@4'

        extractions_with_flag_0 = regexp.extract(test_str, 0, MatchMode.MATCH)
        extractions_with_flag_5 = regexp.extract(test_str, 5, MatchMode.MATCH)

        res_with_flag_0 = [ex.value for ex in extractions_with_flag_0]
        res_with_flag_5 = [ex.value for ex in extractions_with_flag_5]

        expected_res_with_flag_0 = ['a', '1']
        expected_res_with_flag_5 = ['b', '2']
        self.assertEqual(res_with_flag_0, expected_res_with_flag_0)
        self.assertEqual(res_with_flag_5, expected_res_with_flag_5)

    def test_search_mode_with_group(self) -> None:
        regexp = RegexExtractor('(.)@(.)', 'test_extractor')
        test_str = 'testtesttest, a@1, b@2, c@3, d@4'

        extractions_with_flag_0 = regexp.extract(test_str, 0, MatchMode.SEARCH)
        extractions_with_flag_18 = regexp.extract(test_str, 18, MatchMode.SEARCH)

        res_with_flag_0 = [ex.value for ex in extractions_with_flag_0]
        res_with_flag_18 = [ex.value for ex in extractions_with_flag_18]

        expected_res_with_flag_0 = ['a', '1']
        expected_res_with_flag_18 = ['b', '2']
        self.assertEqual(res_with_flag_0, expected_res_with_flag_0)
        self.assertEqual(res_with_flag_18, expected_res_with_flag_18)

    def test_match_mode_without_group(self) -> None:
        regexp = RegexExtractor('.@.', 'test_extractor')
        test_str = 'a@1, b@2, c@3, d@4'

        extractions_with_flag_0 = regexp.extract(test_str, 0, MatchMode.MATCH)
        extractions_with_flag_5 = regexp.extract(test_str, 5, MatchMode.MATCH)

        res_with_flag_0 = [ex.value for ex in extractions_with_flag_0]
        res_with_flag_5 = [ex.value for ex in extractions_with_flag_5]

        expected_res_with_flag_0 = ['a@1']
        expected_res_with_flag_5 = ['b@2']
        self.assertEqual(res_with_flag_0, expected_res_with_flag_0)
        self.assertEqual(res_with_flag_5, expected_res_with_flag_5)

    def test_search_mode_without_group(self) -> None:
        regexp = RegexExtractor('.@.', 'test_extractor')
        test_str = 'testtesttest, a@1, b@2, c@3, d@4'

        extractions_with_flag_0 = regexp.extract(test_str, 0, MatchMode.SEARCH)
        extractions_with_flag_18 = regexp.extract(test_str, 18, MatchMode.SEARCH)

        res_with_flag_0 = [ex.value for ex in extractions_with_flag_0]
        res_with_flag_18 = [ex.value for ex in extractions_with_flag_18]

        expected_res_with_flag_0 = ['a@1']
        expected_res_with_flag_18 = ['b@2']
        self.assertEqual(res_with_flag_0, expected_res_with_flag_0)
        self.assertEqual(res_with_flag_18, expected_res_with_flag_18)

    def test_findall_mode_with_group(self) -> None:
        regexp = RegexExtractor('(.)@(.)', 'test_extractor')
        test_str = 'a@1, b@2, c@3, d@4'

        extractions_with_flag_0 = regexp.extract(test_str, 0, MatchMode.FINDALL)
        extractions_with_flag_5 = regexp.extract(test_str, 5, MatchMode.FINDALL)

        res_with_flag_0 = [ex.value for ex in extractions_with_flag_0]
        res_with_flag_5 = [ex.value for ex in extractions_with_flag_5]

        expected_res_with_flag_0 = ['a', '1', 'b', '2', 'c', '3', 'd', '4']
        expected_res_with_flag_5 = ['b', '2', 'c', '3', 'd', '4']

        self.assertEqual(res_with_flag_0, expected_res_with_flag_0)
        self.assertEqual(res_with_flag_5, expected_res_with_flag_5)

    def test_findall_mode_without_group(self) -> None:
        regexp = RegexExtractor('.@.', 'test_extractor')
        test_str = 'a@1, b@2, c@3, d@4'

        extractions_with_flag_0 = regexp.extract(test_str, 0, MatchMode.FINDALL)
        extractions_with_flag_5 = regexp.extract(test_str, 5, MatchMode.FINDALL)

        res_with_flag_0 = [ex.value for ex in extractions_with_flag_0]
        res_with_flag_5 = [ex.value for ex in extractions_with_flag_5]

        expected_res_with_flag_0 = ['a@1', 'b@2', 'c@3', 'd@4']
        expected_res_with_flag_5 = ['b@2', 'c@3', 'd@4']

        self.assertEqual(res_with_flag_0, expected_res_with_flag_0)
        self.assertEqual(res_with_flag_5, expected_res_with_flag_5)

    def test_split_mode(self) -> None:
        regexp = RegexExtractor(',', 'test_extractor')
        test_str = 'a@1, b@2, c@3, d@4'

        extractions_with_flag_0 = regexp.extract(test_str, 0, MatchMode.SPLIT)
        extractions_with_flag_2 = regexp.extract(test_str, 2, MatchMode.SPLIT)

        res_with_flag_0 = [ex.value for ex in extractions_with_flag_0]
        res_with_flag_2 = [ex.value for ex in extractions_with_flag_2]

        expected_res_with_flag_0 = ['a@1', ' b@2', ' c@3', ' d@4']
        expected_res_with_flag_2 = ['a@1', ' b@2', ' c@3, d@4']

        self.assertEqual(res_with_flag_0, expected_res_with_flag_0)
        self.assertEqual(res_with_flag_2, expected_res_with_flag_2)



import unittest, json
from etk.extractors.email_extractor import EmailExtractor
from etk.etk import ETK
from etk.knowledge_graph_schema import KGSchema


class TestEmailExtractor(unittest.TestCase):

    def test_EmailExtractor(self) -> None:
        kg_schema = KGSchema(json.load(open('etk/unit_tests/ground_truth/test_config.json')))

        etk = ETK(kg_schema=kg_schema, use_spacy_tokenizer=True)

        text = "runqisha@usc.edu 231.45.se.ere@gmail.com " \
               "3233@no.s nefwe.de-ef@l.net  E-mail:anna@wanadoo.fr.com office@fantasie-escort.at"

        email_extractor = EmailExtractor(
            nlp=etk.default_nlp,
            tokenizer=etk.default_tokenizer,
            extractor_name="email_extractor"
        )

        extractions = email_extractor.extract(text)

        extracted = []
        for i in extractions:
            extracted_value = {
                "value": i.value,
                "start_char": i.provenance["start_char"],
                "end_char": i.provenance["end_char"],
                "value_from_text": text[i.provenance["start_char"]: i.provenance["end_char"]]
            }
            extracted.append(extracted_value)
            self.assertEqual(extracted_value["value"], extracted_value["value_from_text"])

        expected = [{'value': 'office@fantasie-escort.at', 'start_char': 97, 'end_char': 122,
                     'value_from_text': 'office@fantasie-escort.at'},
                    {'value': 'runqisha@usc.edu', 'start_char': 0, 'end_char': 16,
                     'value_from_text': 'runqisha@usc.edu'},
                    {'value': 'anna@wanadoo.fr.com', 'start_char': 77, 'end_char': 96,
                     'value_from_text': 'anna@wanadoo.fr.com'},
                    {'value': '231.45.se.ere@gmail.com', 'start_char': 17, 'end_char': 40,
                     'value_from_text': '231.45.se.ere@gmail.com'},
                    {'value': 'nefwe.de-ef@l.net', 'start_char': 51, 'end_char': 68,
                     'value_from_text': 'nefwe.de-ef@l.net'}]

        self.assertEqual(sorted(expected, key=lambda x: x["start_char"]),
                         sorted(extracted, key=lambda x: x["start_char"]))


import re
import etk.timeseries.location_range as lr
#import etk.timeseries.LocationRange as lr

class LocationParser(object):
    def __init__(self):
        self.patterns = {
            'range': re.compile('\[(.+?)\]'),
            'cell_coord': re.compile('\(([A-Z]+),\s*(\d+)\)'),
            'row_label': re.compile('\d+'),
            'col_label': re.compile('[A-Z]+'),
        }

    def parse_range(self, data):
        match = self.patterns['range'].match(data)
        rnge = lr.LocationRange()
        if match:
            for val in [x.strip() for x in match.group(1).split(',')]:
                parts = val.split(':')
                translate_fn = self.translate_row_label
                if self.is_column(parts[0]):
                    translate_fn = self.translate_col_label

                if len(parts) == 1:
                        rnge.add_component(lr.LocationRangeSingletonComponent(translate_fn(parts[0])-1))

                elif len(parts) == 2:
                    if self.is_sentinel(parts[1]):
                        rnge.add_component(lr.LocationRangeInfiniteIntervalComponent(start=translate_fn(parts[0])-1))
                    else:
                        rnge.add_component(lr.LocationRangeIntervalComponent(start=translate_fn(parts[0])-1,
                                                                             end=translate_fn(parts[1])))

                elif len(parts) == 3:

                    if self.is_sentinel(parts[1]):
                        rnge.add_component(lr.LocationRangeInfiniteIntervalComponent(start=translate_fn(parts[0])-1,
                                                                                     increment=int(parts[1])))
                    else:
                        rnge.add_component(lr.LocationRangeIntervalComponent(start=translate_fn(parts[0])-1,
                                                                             end=translate_fn(parts[2]),
                                                                             increment=int(parts[1])))

                else:
                    raise Exception("Error parsing element of range: %s"%val)
        else:
            raise Exception('Error parsing range, data %s did not match pattern'%data)

        return rnge

    def parse_coords(self, data):
        match = self.patterns['cell_coord'].match(data)
        if match:
            return (int(match.group(2)) - 1, self.translate_col_label(match.group(1)) - 1)
        else:
            raise Exception('Error passing cell coordinates: %s'%data)

    def is_column(self, data):
        return self.patterns['col_label'].match(data)

    def is_sentinel(self, data):
        return data == '*'

    def translate_col_label(self, col_label):
        cl = col_label
        if len(cl) == 2:
            return 26 * (ord(cl[0]) - ord('A') + 1) + (ord(cl[1]) - ord('A') + 1)
        elif len(cl) == 1:
            return ord(cl[0]) - ord('A') + 1
        else:
            raise Exception("Unexpected column label: %s"%cl)

    def translate_row_label(self, row_label):
        return int(row_label)



from etk.extractor import Extractor, InputType
from etk.extraction import Extraction
from spacy.matcher import Matcher
from spacy.attrs import LIKE_EMAIL
from typing import List
import copy


FILTER_PROVIDER = ["noon", "no"]


class EmailExtractor(Extractor):
    """
    **Description**
        This class uses spaCy Matcher and takes spaCy predefined 'LIKE_EMAIL' pattern to extract email address.
        More information: https://spacy.io/api/matcher#add

    Examples:
        ::

            email_extractor = EmailExtractor(...)
            email_extractor.extract(text=input_doc,...)

    """
    def __init__(self,
                 nlp,
                 tokenizer,
                 extractor_name: str) -> None:
        """
        Initialize the extractor, storing the rule information and construct spacy rules
        Args:
            nlp:
            tokenizer: Tokenizer
            extractor_name: str

        Returns:
        """
        Extractor.__init__(self,
                           input_type=InputType.TEXT,
                           category="build_in_extractor",
                           name=extractor_name)

        self._nlp = copy.deepcopy(nlp)
        self._like_email_matcher = Matcher(self._nlp.vocab)
        self._tokenizer = tokenizer

    def _load_email_matcher(self):
        self._like_email_matcher.add("Email", None, [{LIKE_EMAIL: True}])

    def extract(self, text: str) -> List[Extraction]:
        """

        Args:
            text (str): The input source to be processed

        Returns:
            List[Extraction]: The list of extractions returned by EmailExtractor

        """

        result = []
        first_phase_doc = self._nlp(text)
        self._load_email_matcher()
        like_email_matches = self._like_email_matcher(first_phase_doc)

        like_emails_filtered = []
        for match_id, start, end in like_email_matches:
            span = first_phase_doc[start:end]
            if self._check_domain(self._tokenizer.tokenize(span.text)):
                like_emails_filtered.append((span.text, span[0].idx, span[-1].idx + len(span[-1])))

        non_space_emails = self._get_non_space_email(first_phase_doc)

        emails = set(like_emails_filtered).union(non_space_emails)

        for email in emails:
            result.append(Extraction(
                value=email[0],
                extractor_name=self.name,
                start_char=email[1],
                end_char=email[2])
            )

        return result

    @staticmethod
    def _check_domain(tokens) -> bool:
        """
        Check if the email provider should be filtered
        Args:
            tokens:

        Returns: Bool
        """

        idx = None
        for e in tokens:
            if e.text == "@":
                idx = e.i
                break
        if not idx or tokens[idx+1].text in FILTER_PROVIDER:
            return False
        else:
            return True

    def _get_non_space_email(self, doc) -> List:
        """
        Deal with corner case that there is "email" string in text and no space around it
        Args:
            doc: List[Token]

        Returns: Bool
        """
        result_lst = []
        for e in doc:
            if "mail:" in e.text.lower():
                idx = e.text.lower().index("mail:") + 5
                value = e.text[idx:]
                tmp_doc = self._nlp(value)
                tmp_email_matches = self._like_email_matcher(tmp_doc)
                for match_id, start, end in tmp_email_matches:
                    span = tmp_doc[start:end]
                    if self._check_domain(self._tokenizer.tokenize(span.text)):
                        result_lst.append((span.text, idx+e.idx, idx+e.idx+len(value)))

        return result_lst



import re
from typing import Optional
from rdflib.namespace import OWL, Namespace, NamespaceManager
from rdflib import URIRef


SCHEMA = Namespace('http://schema.org/')
DIG = Namespace('http://dig.isi.edu/ontologies/dig/')

URI_PATTERN = re.compile(r'^http:|^urn:|^info:|^ftp:|^https:')
URI_ABBR_PATTERN = re.compile(r'^(?:([^:]*):)?([^:]+)$')


class WrongFormatURIException(Exception):
    pass


class PrefixNotFoundException(Exception):
    pass


class PrefixAlreadyUsedException(Exception):
    pass


class OntologyNamespaceManager(NamespaceManager):
    def __init__(self, *args, **kwargs):
        super(OntologyNamespaceManager, self).__init__(*args, **kwargs)
        self.graph.namespace_manager = self
        self.bind('owl', OWL)
        self.bind('schema', SCHEMA)
        self.bind('dig', DIG)

    def parse_uri(self, text: str) -> URIRef:
        """
        Parse input text into URI

        :param text: can be one of
              1. URI, directly return
              2. prefix:name, query namespace for prefix, return expanded URI
              3. name, use default namespace to expand it and return it
        :return: URIRef
        """
        if self.check_uriref(text):
            return self.check_uriref(text)
        elif isinstance(text, str):
            text = text.strip()
            m = URI_ABBR_PATTERN.match(text)
            if m:
                prefix, name = m.groups()
                base = self.store.namespace(prefix if prefix else '')
                if not base:
                    raise PrefixNotFoundException("Prefix: %s", prefix)
                return URIRef(base + name)
        raise WrongFormatURIException()

    def bind(self, prefix: str, namespace: str, override=True, replace=False):
        """
        bind a given namespace to the prefix, forbids same prefix with different namespace

        :param prefix:
        :param namespace:
        :param override: if override, rebind, even if the given namespace is already bound to another prefix.
        :param replace: if replace, replace any existing prefix with the new namespace
        """
        namespace = URIRef(str(namespace))
        # When documenting explain that override only applies in what cases
        if prefix is None:
            prefix = ''
        bound_namespace = self.store.namespace(prefix)
        # Check if the bound_namespace contains a URI and if so convert it into a URIRef for
        # comparison. This is to prevent duplicate namespaces with the same URI.
        if bound_namespace:
            bound_namespace = URIRef(bound_namespace)
        if bound_namespace and bound_namespace != namespace:

            if replace:
                self.store.bind(prefix, namespace)
            # prefix already in use for different namespace
            raise PrefixAlreadyUsedException("Prefix (%s, %s) already used, instead of (%s, %s).",
                                             prefix, self.store.namespace(prefix).toPython(),
                                             prefix, namespace.toPython())
        else:
            bound_prefix = self.store.prefix(namespace)
            if bound_prefix is None:
                self.store.bind(prefix, namespace)
            elif bound_prefix == prefix:
                pass  # already bound
            else:
                if override or bound_prefix.startswith("_"):
                    self.store.bind(prefix, namespace)

    @staticmethod
    def check_uriref(text: str) -> Optional[URIRef]:
        if isinstance(text, URIRef):
            return text
        if isinstance(text, str):
            text = text.strip()
            if URI_PATTERN.match(text.strip()):
                return URIRef(text)



import unittest, json
from etk.etk import ETK
from etk.extractors.glossary_extractor import GlossaryExtractor
from etk.knowledge_graph_schema import KGSchema

sample_input = {
        "projects": [
            {
                "name": "etk",
                "description": "version 2 of etk, implemented by Runqi, Dongyu, Sylvia, Amandeep and others."
            },
            {
                "name": "rltk",
                "description": "record linkage toolkit, implemented by Pedro, Mayank, Yixiang and several students."
            }
        ]
    }


class TestProvenance(unittest.TestCase):

    def test_Provenance(self) -> None:
        kg_schema = KGSchema(json.load(open('etk/unit_tests/ground_truth/test_config.json')))

        self.etk = ETK(kg_schema=kg_schema, use_spacy_tokenizer=True)
        g = ['runqi', 'sylvia', 'dongyu', 'mayank', 'pedro', 'amandeep', 'yixiang']
        self.name_extractor = GlossaryExtractor(g, "name_extractor",
                                                self.etk.default_tokenizer,
                                                case_sensitive=False, ngrams=1)
        doc = self.etk.create_document(sample_input)
        descriptions = doc.select_segments("projects[*].description")
        projects = doc.select_segments("projects[*]")

        for d, p in zip(descriptions, projects):
            names = doc.extract(self.name_extractor, d)
            p.store(names, "members")

        expected_provenances = [
            {
              "@id": 0,
              "@type": "extraction_provenance_record",
              "method": "name_extractor",
              "confidence": 1.0,
              "origin_record": {
                "path": "projects.[0].description",
                "start_char": 33,
                "end_char": 38
              }
            },
            {
              "@id": 1,
              "@type": "extraction_provenance_record",
              "method": "name_extractor",
              "confidence": 1.0,
              "origin_record": {
                "path": "projects.[0].description",
                "start_char": 40,
                "end_char": 46
              }
            },
            {
              "@id": 2,
              "@type": "extraction_provenance_record",
              "method": "name_extractor",
              "confidence": 1.0,
              "origin_record": {
                "path": "projects.[0].description",
                "start_char": 48,
                "end_char": 54
              }
            },
            {
              "@id": 3,
              "@type": "extraction_provenance_record",
              "method": "name_extractor",
              "confidence": 1.0,
              "origin_record": {
                "path": "projects.[0].description",
                "start_char": 56,
                "end_char": 64
              }
            },
            {
              "@id": 4,
              "@type": "storage_provenance_record",
              "doc_id": None,
              "field": None,
              "destination": "projects.[0].members",
              "parent_provenances": {
                "Runqi": 0,
                "Dongyu": 1,
                "Sylvia": 2,
                "Amandeep": 3
              }
            },
            {
              "@id": 5,
              "@type": "extraction_provenance_record",
              "method": "name_extractor",
              "confidence": 1.0,
              "origin_record": {
                "path": "projects.[1].description",
                "start_char": 39,
                "end_char": 44
              }
            },
            {
              "@id": 6,
              "@type": "extraction_provenance_record",
              "method": "name_extractor",
              "confidence": 1.0,
              "origin_record": {
                "path": "projects.[1].description",
                "start_char": 46,
                "end_char": 52
              }
            },
            {
              "@id": 7,
              "@type": "extraction_provenance_record",
              "method": "name_extractor",
              "confidence": 1.0,
              "origin_record": {
                "path": "projects.[1].description",
                "start_char": 54,
                "end_char": 61
              }
            },
            {
              "@id": 8,
              "@type": "storage_provenance_record",
              "doc_id": None,
              "field": None,
              "destination": "projects.[1].members",
              "parent_provenances": {
                "Pedro": 5,
                "Mayank": 6,
                "Yixiang": 7
              }
            }
            ]
        expected_projects = [
                {
                  "name": "etk",
                  "description": "version 2 of etk, implemented by Runqi, Dongyu, Sylvia, Amandeep and others.",
                  "members": [
                    "Runqi",
                    "Dongyu",
                    "Sylvia",
                    "Amandeep"
                  ]
                },
                {
                  "name": "rltk",
                  "description": "record linkage toolkit, implemented by Pedro, Mayank, Yixiang and several students.",
                  "members": [
                    "Pedro",
                    "Mayank",
                    "Yixiang"
                  ]
                }
              ]
        self.assertEqual(expected_projects, doc.value["projects"])
        self.assertEqual(expected_provenances, doc.value["provenances"])

if __name__ == '__main__':
    unittest.main()



from typing import List
from etk.extraction import Extraction


class BlackListFilter(object):
    def __init__(self, black_list: List[str]) -> None:
        self.black_list = black_list

    def filter(self, extractions, case_sensitive=False) -> List[Extraction]:
        """filters out the extraction if extracted value is in the blacklist"""
        filtered_extractions = []
        if not isinstance(extractions, list):
            extractions = [extractions]

        for extraction in extractions:
            if case_sensitive:
                try:
                    if extraction.value.lower() not in self.black_list:
                        filtered_extractions.append(extraction)
                except Exception as e:
                    print('Error in BlackListFilter: {} while filtering out extraction: {}'.format(e, extraction.value))
                    # most likely it s a unicode character which is messing things up, return it
                    filtered_extractions.append(extraction)
            else:
                if extraction.value not in self.black_list:
                    filtered_extractions.append(extraction)
        return filtered_extractions



from typing import List
from etk.extractor import Extractor, InputType
from etk.extraction import Extraction, Extractable
from extruct.w3cmicrodata import MicrodataExtractor
from extruct.jsonld import JsonLdExtractor
from extruct.rdfa import RDFaExtractor
from bs4 import BeautifulSoup


class HTMLMetadataExtractor(Extractor):
    """
    **Description**
        Extracts META, microdata, JSON-LD and RDFa from HTML pages.

        Uses https://stackoverflow.com/questions/36768068/get-meta-tag-content-property-with-beautifulsoup-and-python to
        extract the META tags

        Uses https://github.com/scrapinghub/extruct to extract metadata from HTML pages

    Examples:
        ::

            html_metadata_extractor = HTMLMetadataExtractor()
            html_metadata_extractor.extract(text=input_doc,
                                            extract_title=True,
                                            extract_meta=True,
                                            extract_microdata=False,
                                            extract_json_ld=False,
                                            extract_rdfa=False)

    """

    def __init__(self):
        Extractor.__init__(self,
                           input_type=InputType.HTML,
                           category="HTML extractor",
                           name="HTML metadata extractor")

    def extract(self, html_text: str,
                extract_title: bool = False,
                extract_meta: bool = False,
                extract_microdata: bool = False,
                extract_json_ld: bool = False,
                extract_rdfa: bool = False) \
            -> List[Extraction]:
        """
        Args:
            html_text (str): input html string to be extracted
            extract_title (bool): True if string of 'title' tag needs to be extracted, return as { "title": "..." }
            extract_meta (bool): True if string of 'meta' tags needs to be extracted, return as { "meta": { "author": "...", ...}}
            extract_microdata (bool): True if microdata needs to be extracted, returns as { "microdata": [...] }
            extract_json_ld (bool): True if json-ld needs to be extracted, return as { "json-ld": [...] }
            extract_rdfa (bool): True if rdfs needs to be extracted, returns as { "rdfa": [...] }

        Returns:
            List[Extraction]: the list of extraction or the empty list if there are no matches.
        """
        res = list()
        soup = BeautifulSoup(html_text, 'html.parser')

        if soup.title and extract_title:
            title = self._wrap_data("title", soup.title.string.encode('utf-8').decode('utf-8'))
            res.append(title)

        if soup.title and extract_meta:
            meta_content = self._wrap_meta_content(soup.find_all("meta"))
            meta_data = self._wrap_data("meta", meta_content)
            res.append(meta_data)

        if extract_microdata:
            mde = MicrodataExtractor()
            mde_data = self._wrap_data("microdata", mde.extract(html_text))
            res.append(mde_data)

        if extract_json_ld:
            jslde = JsonLdExtractor()
            jslde_data = self._wrap_data("json-ld", jslde.extract(html_text))
            res.append(jslde_data)

        if extract_rdfa:
            rdfae = RDFaExtractor()
            rdfae_data = self._wrap_data("rdfa", rdfae.extract(html_text))
            res.append(rdfae_data)

        return res

    def _wrap_data(self, key: str, value) -> Extraction:
        e = Extraction(value=value, extractor_name=self.name, tag=key)
        return e

    @staticmethod
    def _wrap_meta_content(meta_tags) -> dict:
        meta = {}
        for tag in meta_tags:
            meta[tag.get("name")] = tag.get("content")

        return meta



import json
from typing import List
from etk.extractor import Extractor, InputType
from etk.extraction import Extraction
from etk.etk_exceptions import ExtractorError
from etk.dependencies.landmark.landmark_extractor.extraction.Landmark import ItemRule, IterationRule, loadRule


class InferlinkRule(object):
    """
    Wrapper of a single rule.
    """

    def __init__(self, rule: dict):
        self._name = rule['name'] if 'name' in rule else ''
        self._value = None
        self._start_char = None
        self._end_char = None
        self._rule = loadRule(rule)

    @property
    def name(self) -> str:
        return self._name

    @property
    def value(self) -> str:
        return self._value

    @property
    def start_char(self) -> int:
        return self._start_char

    @property
    def end_char(self) -> int:
        return self._end_char

    def apply(self, html_text: str):
        extraction = self._rule.apply(html_text)  # a dict with 'rule_id', 'extract', 'begin_index', 'end_index'...
        self._value = extraction['extract'] if 'extract' in extraction and extraction['extract'] != '' else None
        self._start_char = extraction['begin_index'] if 'begin_index' in extraction and extraction[
            'extract'] != -1 else None
        self._end_char = extraction['end_index'] if 'end_index' in extraction and extraction['extract'] != -1 else None


class InferlinkRuleSet(object):
    """
    Wrapper class on an inferlink JSON to provide a convenient API to work with rules.
    """

    def __init__(self, rule_set: List[dict]):
        self._rules = [InferlinkRule(rule) for rule in rule_set]

    @property
    def rules(self) -> List[InferlinkRule]:
        return self._rules

    @staticmethod
    def load_rules_file(file_name: str) -> List[dict]:
        with open(file_name, 'r') as f:
            return json.load(f)['rules']


class InferlinkExtractor(Extractor):
    """
    **Description**
        This class extracts segments from an HTML page using rules created by the Inferlink web wrapper.

    Examples:
        ::

            inferlink_extractor = InferlinkExtractor()
            inferlink_extractor.extract(text=input_doc,
                                        threshold=0.8)
    """

    def __init__(self, rule_set: InferlinkRuleSet):
        Extractor.__init__(self,
                           input_type=InputType.HTML,
                           category="HTML extractor",
                           name="Inferlink extractor")
        self._rule_set = rule_set

    def extract(self, html_text: str, threshold=0.5) -> List[Extraction]:
        """

        Args:
            html_text (str): str of the html page to be extracted
            threshold (float): if the ratio of rules that successfully extracted something over all rules \
                    is higher than or equal to the threshold, return the results, else return an empty list

        Returns:
            List[Extraction]: a list of Extractions, each extraction includes the extracted value, the rule name, the provenance etc.

        """

        result = list()
        try:
            for rule in self._rule_set.rules:
                rule.apply(html_text)
                value = rule.value
                if value is not None:
                    # note the addition of a new tag argument to Extraction
                    start_char = rule.start_char
                    end_char = rule.end_char
                    result.append(Extraction(value, self.name, start_char=start_char, end_char=end_char, tag=rule.name))

            # Test whether the fraction of extractions meets the desired threshold
            if len(self._rule_set.rules) > 0 and float(len(result)) / len(self._rule_set.rules) >= threshold:
                return result
            else:
                return list()
        except Exception as e:
            raise ExtractorError('Error in extracting landmark %s' % e)



import unittest, json
from etk.knowledge_graph import KGSchema
from etk.etk import ETK
from etk.etk_exceptions import KgValueError
from datetime import date, datetime
from etk.provenance_api import ProvenanceAPI


class TestKnowledgeGraphProvenance(unittest.TestCase):

    def test_KnowledgeGraph_provenance(self) -> None:
        sample_doc = {
            "projects": [
                {
                    "name": "etk",
                    "description": "version 2 of etk, implemented by Runqi12 Shao, Dongyu Li, Sylvia lin, Amandeep and others.",
                    "members": [
                        "dongyu",
                        "amandeep",
                        "sylvia",
                        "Runqi12"
                    ],
                    "date": "2007-12-05",
                    "place": "columbus:georgia:united states:-84.98771:32.46098"
                },
                {
                    "name": "rltk",
                    "description": "record linkage toolkit, implemented by Pedro, Mayank, Yixiang and several students.",
                    "members": [
                        "mayank",
                        "yixiang"
                    ],
                    "date": ["2007-12-05T23:19:00"],
                    "cost": -3213.32
                }
            ]
        }

        kg_schema = KGSchema(json.load(open('etk/unit_tests/ground_truth/test_config.json')))

        etk = ETK(kg_schema)
        doc = etk.create_document(sample_doc)

        try:
            doc.kg.add_value("developer", json_path="projects[*].members[*]")
        except KgValueError:
            pass

        try:
            doc.kg.add_value("test_date", json_path="projects[*].date[*]")
        except KgValueError:
            pass

        try:
            doc.kg.add_value("test_add_value_date", value=[date(2018, 3, 28), {}, datetime(2018, 3, 28, 1, 1, 1)],
                             json_path_extraction="projects[0].date")
        except KgValueError:
            pass

        try:
            doc.kg.add_value("test_location", json_path="projects[*].place")
        except KgValueError:
            pass

        # print (json.dumps(doc.value, indent=2))

        expeced_provenances = [
            {
                "@id": 0,
                "@type": "kg_provenance_record",
                "reference_type": "location",
                "value": "dongyu",
                "json_path": "projects.[0].members.[0]"
            },
            {
                "@id": 1,
                "@type": "kg_provenance_record",
                "reference_type": "location",
                "value": "amandeep",
                "json_path": "projects.[0].members.[1]"
            },
            {
                "@id": 2,
                "@type": "kg_provenance_record",
                "reference_type": "location",
                "value": "sylvia",
                "json_path": "projects.[0].members.[2]"
            },
            {
                "@id": 3,
                "@type": "kg_provenance_record",
                "reference_type": "location",
                "value": "Runqi12",
                "json_path": "projects.[0].members.[3]"
            },
            {
                "@id": 4,
                "@type": "kg_provenance_record",
                "reference_type": "location",
                "value": "mayank",
                "json_path": "projects.[1].members.[0]"
            },
            {
                "@id": 5,
                "@type": "kg_provenance_record",
                "reference_type": "location",
                "value": "yixiang",
                "json_path": "projects.[1].members.[1]"
            },
            {
                "@id": 6,
                "@type": "kg_provenance_record",
                "reference_type": "location",
                "value": "2007-12-05T00:00:00",
                "json_path": "projects.[0].date.[0]"
            },
            {
                "@id": 7,
                "@type": "kg_provenance_record",
                "reference_type": "location",
                "value": "2007-12-05T23:19:00",
                "json_path": "projects.[1].date.[0]"
            },
            {
                "@id": 8,
                "@type": "kg_provenance_record",
                "reference_type": "constant",
                "value": "2018-03-28",
                "json_path": "projects[0].date"
            },
            {
                "@id": 9,
                "@type": "kg_provenance_record",
                "reference_type": "constant",
                "value": "2018-03-28T01:01:01",
                "json_path": "projects[0].date"
            },
            {
                "@id": 10,
                "@type": "kg_provenance_record",
                "reference_type": "location",
                "value": "columbus:georgia:united states:-84.98771:32.46098",
                "json_path": "projects.[0].place"
            }
        ]

        self.assertEqual(expeced_provenances, doc.value["provenances"])



from typing import List
from enum import Enum, auto
from etk.extractor import Extractor, InputType
from etk.extraction import Extraction
import re
import collections


class MatchMode(Enum):
    MATCH = auto(),
    SEARCH = auto(),
    FINDALL = auto(),
    SPLIT = auto()


class RegexExtractor(Extractor):
    """
    **Description**
        A wrapper for Python regular expressions.

    Examples:
        ::

            pattern = "some_pattern"
            regex_extractor = RegexExtractor(pattern=pattern,
                                            flags=re.IGNORECASE)
            regex_extractor.extract(text=input_doc,
                                    flags=re.M,
                                    MatchMode=MatchMode.SEARCH
                                    )

    """

    def __init__(self,
                 pattern: str,
                 extractor_name: str='regex extractor',
                 flags=0,
                 general_tag: str=None
                 ) -> None:
        Extractor.__init__(self,
                           input_type=InputType.TEXT,
                           category="regex",
                           name=extractor_name)

        self._compiled_regex = re.compile(pattern, flags)
        self._general_tag = general_tag

        self._match_functions = {
            MatchMode.MATCH: self._compiled_regex.match,
            MatchMode.SEARCH: self._compiled_regex.search,
            MatchMode.FINDALL: self._compiled_regex.finditer,
            MatchMode.SPLIT: self._compiled_regex.split
        }

    @property
    def general_tag(self):
        return self._general_tag

    def extract(self, text: str, flags=0, mode: MatchMode = MatchMode.FINDALL) -> List[Extraction]:

        """

            Extracts information from a text using the given regex.
            If the pattern has no groups, it returns a list with a single Extraction.
            If the pattern has groups, it returns a list of Extraction, one for each group.
            Each extraction records the start and end char positions of matches.

        Args:
            text (str): the text to extract from.
            flags (enum['a', 'i', 'L', 'm', 's', 'u', 'x']): flags given to search or match. The value should be one \
                or more letters from the set 'a', 'i', 'L', 'm', 's', 'u', 'x'.) The group matches the empty string; \
                the letters set the corresponding flags: re.A (ASCII-only matching), re.I (ignore case), re.L (locale dependent),\
                re.M (multi-line), re.S (dot matches all), re.U (Unicode matching), and re.X (verbose), for the entire \
                regular expression.
            mode (enum[MatchMode.MATCH, MatchMode.SEARCH, MatchMode.FINDALL, MatchMode.SPLIT]): whether to use re.search() or re.match().

        Returns:
            List(Extraction): the list of extraction or the empty list if there are no matches.

        """

        match_func = self._match_functions[mode]
        matches = match_func(text, flags)
        return self._wrap_result(matches)

    # wrap the re return object to list of extraction
    def _wrap_result(self, matches: object) -> List[Extraction]:
        res = list()
        # matches are result of split()
        if isinstance(matches, list):
            return self._wrap_split_extraction(matches)

        # matches are result of finditer()
        elif isinstance(matches, collections.Iterable):
            for match in matches:
                es = self._wrap_result(match)
                res.extend(es)

        # single match
        else:
            # check if the pattern has groups
            groups = matches.groups()
            if groups:
                for i in range(1, len(groups) + 1):
                    res.append(self._wrap_extraction(i, matches))
            else:
                res.append(self._wrap_extraction(0, matches))
        return res

    def _wrap_split_extraction(self, items: List[str]) -> List[Extraction]:
        res = list()
        start = 0
        for item in items:
            end = start + len(item)
            e = Extraction(value=item, extractor_name=self.name, start_char=start, end_char=end)
            res.append(e)
            start = end
        return res

    def _wrap_extraction(self, group_idx: int, matches: object) -> Extraction:
        start, end = matches.start(group_idx), matches.end(group_idx)
        text = matches.group(group_idx)
        e = Extraction(value=text, extractor_name=self.name, \
                       start_char=start, end_char=end, tag=self.general_tag)
        return e
